#include <map.h>

const double PIXEL_WIDTH = 609;
const double PIXEL_HEIGHT = 354;


bool Map_Connection::create_image(const RectHull& rh)
{
    mapObj* map = NULL;
    imageObj* msImage = NULL;

    std::string mapFile = "/occ/msl/web/90/001/mapfile/test3.map";
    // Load the config file
    map = msLoadMap(const_cast<char*>(mapFile.c_str()), NULL);
    if (map == NULL)
    {
        report("Map_Connection::create_image", "Failed to load map file \"%s\"", Error::ERROR, __LINE__, mapFile.c_str());
        msWriteError(stderr);
        return false;
    }

    // Make sure it generates a jpg
    outputFormatObj *format;
    format = msSelectOutputFormat(map, "jpeg");
    if (format == NULL)
    {
        report("Map_Connection::create_image", "Failed create format", Error::ERROR, __LINE__);
        msWriteError(stderr);
        msFreeMap(map);
        msCleanup();
        return false;
    }
    msFree((char *) map->imagetype);
    map->imagetype = strdup("jpeg");
    msApplyOutputFormat(&(map->outputformat), format, MS_NOOVERRIDE, MS_NOOVERRIDE, MS_NOOVERRIDE);

    // Set the image size
    map->height = static_cast<int>(PIXEL_HEIGHT);
    map->width = static_cast<int>(PIXEL_WIDTH);

    // Allow non-square pixels
    msSetConfigOption(map, "MS_NONSQUARE", "YES");

    // Set the Rect Hull of the image
    map->extent.minx = rh[SW].x();
    map->extent.maxx = rh[SE].x();
    map->extent.miny = rh[SW].y();
    map->extent.maxy = rh[NW].y();

    // Create the image
    if (msImage)
    {
        delete msImage;
        msImage = NULL;
    }
// Save the mapfile before calling msDrawMap()
std::string fileName2 = "/tmp/before_";
fileName2 += ntos((long)brianX);
fileName2 += "_";
fileName2 += ntos((long)brianY);
fileName2 += ".out";
msSaveMap(map, (char*)fileName2.c_str());

    msImage = msDrawMap(map);

// Save the mapfile after calling msDrawMap()
std::string fileName3 = "/tmp/after_";
fileName3 += ntos((long)brianX);
fileName3 += "_";
fileName3 += ntos((long)brianY);
fileName3 += ".out";
msSaveMap(map, (char*)fileName3.c_str());
report("Map_Connection::create_image", " (%d %d) Map scale is %lf, calc scale is %lf", Error::ERROR, __LINE__, brianX, brianY, map->scale, _scale);

    if (msImage == NULL)
    {
        report("Map_Connection::create_image", "Failed create image", Error::ERROR, __LINE__);
        msWriteError(stderr);
        msFreeMap(map);
        msCleanup();
        return false;
    }

    // Reset the image
    _image_ptr = gdImageJpegPtr( msImage->img.gd, &_image_size, -1 );

    _x_pixels = map->width;
    _y_pixels = map->height;

    msFreeMap(map);
    msCleanup();

std::string fileName = "/tmp/new_";

fileName += ntos((long)brianX);
fileName += "_";
fileName += ntos((long)brianY);
fileName += ".jpg";
std::ofstream ofs(fileName.c_str());
ofs.write( (const char *)_image_ptr, _image_size );
ofs.close();

    return true;
}

void Map_Connection::setupImageInfo(DrawSetupInfo &info, const RectHull & pictureHull )
{
    FeatureDef::Type ft;
    std::string fn, plotColorValue;

    ColorTxtSet colorSet;
    ColorTxt colorTxt;

    initColorTxtSet(colorSet);

    // calculate size of picture hull in feet -- we recalculate the
    // image width and height for the DrawingDevice here so that the
    // image hull and output page are perfectly aligned.

    double minLatitude = std::min(fabs(pictureHull[NW].y()), fabs(pictureHull[SW].y()));
    double picWidth = fabs(Vertex::FEET_PER_MILE *
                    ((Vertex::MILES_PER_DEGREE * pictureHull[NE].x()) -
                     (Vertex::MILES_PER_DEGREE * pictureHull[NW].x())) *
                    cos(degreesToRadians(minLatitude)));
    double picHeight = fabs(Vertex::FEET_PER_MILE *
                  ((Vertex::MILES_PER_DEGREE * pictureHull[NW].y()) -
                   (Vertex::MILES_PER_DEGREE * pictureHull[SW].y())) );

    info.width = picWidth / _scale;
    info.height = picHeight / _scale;

    FeatureTypeAllLoop(ft)
    {
        if (FeatureDef::getPresent(ft))
        {
            info.ftypeSelected[ft] = FeatureDef::getDisplay(ft);
            info.ftypeLabel[ft] = FeatureDef::getLabel(ft);
        }
        else
        {
            info.ftypeSelected[ft] = false;
            info.ftypeLabel[ft]    = false;
            info.ftypeColor[ft]    = RGB(255, 255, 255); // White
            info.ftypeLabelColor[ft] = RGB(255, 255, 255); // White
        }
        colorTxt.setName(FeatureDef::getColor(ft));
        info.ftypeColor[ft] = *(colorSet.find(colorTxt));
        info.ftypeLabelColor[ft] = *(colorSet.find(colorTxt));
        info.ftypeLineStyle[ft] = FeatureDef::getLineStyle(ft);
        info.ftypeLineWidth[ft] = FeatureDef::getLineWidth(ft);
    }

    // Setup the district polygon color
    colorTxt.setName(FeatureDef::getColor(FeatureDef::DistPoly));
    info.distPolyFeatureColorList.push_back(*(colorSet.find(colorTxt)));

    // Never label Grids, handled by the applet
    info.ftypeLabel[FeatureDef::Grid] = false;
    if (_scale > config->get_grid_scale_override())
    {
        info.ftypeSelected[FeatureDef::Grid] = false;
        info.ftypeLabel[FeatureDef::Grid] = false;
    }

    // Add Transaction Info
    info.transactionSelected = false;
    info.transactionLabel = false;
    // Transactions aren't displayed
    //info.transactionColor = DEFAULT_TRANS_COLOR;
    //info.transactionLineStyle = DEFAULT_TRANS_LINE_STYLE;
    //info.transactionLineWidth = DEFAULT_LINE_WIDTH_STR;

    info.scaleMagnification = _scale;
    info.deviceSelected = DrawSetupInfo::GRAPHIC_DRAW;
    info.outputFileName = "";
    info.comment = "";
    info.distPolySpecFile = "";
    info.distPolySpec = false;
    info.setTransFileNames("");
    info.curCountyOnly = true;
    info.drawFrame = false;
    info.drawMargin = false;
}

// ****************************************************************
// NAME:        Map_Connection::calculateDrawScale
//
// DESCRIPTION: Calculates the current scale of the image
//
// PARAMETERS:  <Parm name> <Parm type>    <Short description>
//              rh          const RectHull&  the area to calculate
//
// RETURNS:     None
//
// ****************************************************************
void Map_Connection::calculateDrawScale(const RectHull& rh)
{
    // Calculate feet per inch of the onscreen map
    _scale = DrawGD::PIXELS_PER_SCREEN_INCH * rh.diagonalGEO() * Vertex::FEET_PER_MILE / sqrt( PIXEL_HEIGHT * PIXEL_HEIGHT + PIXEL_WIDTH * PIXEL_WIDTH);
}

// ****************************************************************
// NAME:        Map_Connection::getDistPolyFeatures
//
// DESCRIPTION: Retrieves the distpoly features for a given RectHull
//
//
// PARAMETERS:  <Parm name>       <Parm type>      <Short description>
//              rh                const RectHull&  the area to find in
//              distPolyFeatures  FeatureList&     the list to store in
//
// RETURNS:     None
//
// ****************************************************************
void Map_Connection::getDistPolyFeatures(const RectHull& rh, FeatureList& distPolyFeatures)
{
    if (! FeatureDef::getDisplay(FeatureDef::DistPoly))
    {
        report("Map_Connection::getDistPolyFeatures", "FeatureDef::DistPoly type disabled from display", Error::DEBUG, __LINE__);
        return;
    }

    OvlQuery query;
    DrawSetupInfo info;

    report("Map_Connection::getDistPolyFeatures", "create image from RectHull RectHull: (%f, %f)-(%f, %f)", Error::DEBUG, __LINE__, rh[NW].x(), rh[NW].y(), rh[SE].x(), rh[SE].y());

    // Need to calculate zoom level
    setupImageInfo(info, rh);

    query.requestState(mapBase->getStateCode());
    query.requestCounty(mapBase->getCountyCode());
    query.requestExtended();
    query.requestHull(rh);
    // enforce zooming
    query.requestDisplayLevel(PageDraw::calculateDisplayLevel(info.scaleMagnification));
    query.requestGeneralize();
    query.requestCombine();

    // This will only retrieve distpoly features
    query.requestType(FeatureDef::DistPoly);

    if (! OvlMgr::getFeatures(query, distPolyFeatures))
        return;
    // Get rid of copied features
    FeatureList tList;
    if (OvlUtil::removeCopies(distPolyFeatures, tList))
        tList.erase(tList.begin(), tList.end());

    report("Map_Connection::getDistPolyFeatures", "found %d features", Error::DEBUG, __LINE__, distPolyFeatures.size());

    return;
}

int Map_Connection::get_intersections( OvlQuery & main_query, OvlQuery & intersect_query, IntersectionPtrList & ilist )
{
    // find the interstection of the main street and the cross street,
    // using a tolerance of 25 feet.
    
    static const double tolerance = 25.0; // feet
        
    if (! OvlMgr::getIntersection( main_query, intersect_query, tolerance, ilist )){
        errRptN(( "OvlMgr::getIntersection failed" ));
        return 0;
    }

    // eliminate duplicate entries in the list
    
    IntersectionPtrList::iterator iter1, iter2;

    for (iter1 = ilist.begin(); iter1 != ilist.end(); iter1++)
    {
        iter2 = iter1;
        iter2++;

        while (iter2 != ilist.end())
        {
            if ((*iter1)->point().sameVertex((*iter2)->point(), Vertex::EpsilonGEO ))
            {
                IntersectionPtrList::iterator eraseiter = iter2;
                iter2++;
                ilist.erase(eraseiter);
            }
            else 
                iter2++;
        }
    }

    return ilist.size();
}


void Map_Connection::get_current_feature_list( IntersectionPtrList & il )
{
    IntersectionPtrList::iterator iter1, iter2;

    for (IntersectionPtrList::iterator iter = il.begin(); iter != il.end(); iter++)
        _current_feature_list.push_back(*((*iter)->secondFeature()));
}


bool Map_Connection::secondary_search( AddressQueryInfo & street, OvlQuery & query, FeatureList & fl )
{
    query.clearMcd();

    if ( make_secondary_search( street, query )){
        if ( OvlMgr::getFeatures( query , fl ) && fl.size() > 0 ){
            return true;
        }
        report("Map_Connection::secondary_search", "no features", Error::DEBUG, __LINE__);
    }
    report("Map_Connection::secondary_search", "secondary search failed", Error::DEBUG, __LINE__);

    return false;
}


bool Map_Connection::primary_search( AddressQueryInfo & street, OvlQuery & query, FeatureList & fl )
{
    if ( get_street_ids( street, query )){
        if ( ! OvlMgr::getFeatures( query , fl ) || fl.size() == 0 ){
            report("Map_Connection::primary_search", "no features", Error::DEBUG, __LINE__);
            return false;
        }
    }
    else {
        return false;
    }
           
    return true;
}


int Map_Connection::trsq_search(bool initialSearch)
{
    GridUtil::TRSQGrid srcGrid;
    GridUtil::TRSQGridVector resultGrids;
    std::string state_name, county_name;

    if (initialSearch)
        failedPrimarySearch = true;

    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights
    state_name = get_value( "state" );
    county_name = get_value( "county" );

    srcGrid.township = get_value("map");
    srcGrid.range = get_value("page");
    srcGrid.name = get_value("grids");

    sutil::uppercase( state_name );
    sutil::uppercase( county_name );
    sutil::uppercase(srcGrid.township);
    sutil::uppercase(srcGrid.range);
    sutil::uppercase(srcGrid.range);
    
    if (srcGrid.township.empty() || srcGrid.range.empty() || srcGrid.name.empty())
    {
        errRptN(( "missing trsq search parameters: Township: \"%s\" Range: \"%s\" Section/Quater \"%s\"", srcGrid.township.c_str(), srcGrid.range.c_str(), srcGrid.name.c_str()));
        return 0;
    }

// TODO: Handle Alberta when they start internet mapping
    GridUtil::format_for_maps(srcGrid, false, resultGrids);

    if (resultGrids.size() == 0)
    {
        errRptN(( "Failed to lookup TRSQ parameters  resultGrids.size(): %d", resultGrids.size()));
        return 0;
    }

    errRptI(( "resultGrids.size(): %d", resultGrids.size()));
    
    report("Map_Connection::trsq_search", "state_name:  %s", Error::DEBUG, __LINE__, state_name.c_str());
    report("Map_Connection::trsq_search", "county_name: %s", Error::DEBUG, __LINE__, county_name.c_str());

    _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
    for (IntersectionPtrList::iterator iter = _current_intersection_list.begin(); iter != _current_intersection_list.end(); iter++)
        delete *iter;
    _current_intersection_list.erase(_current_intersection_list.begin(), _current_intersection_list.end());
    if (!imap::setStateCounty(state_name, county_name, mapBase))
    {
        report("Map_Connection::trsq_search", "Failed to set state/county", Error::ERROR, __LINE__);
        return 0;
    }

    OvlQuery query;
    
    query.requestState( mapBase->getStateCode() );
    query.requestCounty( mapBase->getCountyCode() );
    query.requestExtended();

    for (unsigned int i = 0; i < resultGrids.size(); i++)
    {
        msl::IntList nids;

        errRptI(( "get ids for trsq: %s", resultGrids[i].name.c_str()));
            
        if ( mapBase->getNameManager().getNameIDs(resultGrids[i].name.c_str(), nids ))
        {
            FName fname;
            fname.prefix = fname.suffix = FeatureDef::AnyDirection;
            fname.type = FeatureDef::AnyNameType;
            fname.name = *(nids.begin());
            query.requestName( fname );
        }
        else
        {
            report("Map_Connection::trsq_search", "Invalid grid name \"%s\"", Error::ERROR, __LINE__, resultGrids[i].name.c_str());
            return 0;
        }
    }
        
    query.clearType();

    // FIXME: need to verify if this is the correct feature set to
    // search, or read it from a parm if possible.  Also need to
    // verify/read from a parm that this is the correct feature
    // set to use to set polygon_is_grid flag.
        
    if (FeatureDef::getSearch(FeatureDef::Grid) == true)
        query.requestType( FeatureDef::Grid );
    if (FeatureDef::getSearch(FeatureDef::DistPoly) == true)
        query.requestType( FeatureDef::DistPoly );
            
    if ( OvlMgr::getFeatures( query, _current_feature_list ) && _current_feature_list.size() != 0 )
    {
        Locate locate_poly(mapBase->getStateCode(), mapBase->getCountyCode());
        locate_poly.initLocate(_current_feature_list);
        const RectHull & rh = _hull_manager.reset( locate_poly );
        _hull_manager.set_polygon_is_grid( true );
        create_image(rh);
        if (initialSearch)
           failedPrimarySearch = false;
    }

    report("Map_Connection::trsq_search", "_current_feature_list.size() = %d", Error::DEBUG, __LINE__, _current_feature_list.size());

    return _current_feature_list.size();
}


bool Map_Connection::ll_search(bool initialSearch)
{
    string coord_type;
    double tmpLat, tmpLon, lat, lon;
    bool baseNAD83, requestNAD83;

    if (initialSearch)
        failedPrimarySearch = true;

    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights

    requestNAD83 = (get_value("NAD83") == "true");
    coord_type = get_value("coord_type");
    report("Map_Connection::ll_search", "Coord_type %s NAD83 %d(%s)", Error::DEBUG, __LINE__, coord_type.c_str(), requestNAD83, get_value("NAD83").c_str());

    if  (coord_type.empty() ||
         ! (coord_type == DECIMAL_LL_SEARCH || coord_type == DMS_LL_SEARCH || coord_type == SPCS_LL_SEARCH || coord_type == UTM_LL_SEARCH))
    {
        report("Map_Connection::ll_search", "Missing/invalid coord parm %s", Error::WARNING, __LINE__, coord_type.c_str());
        return false;
    }

    if (!imap::setStateCounty(get_value("STATE"), get_value("COUNTY"), mapBase))
    {
        report("Map_Connection::ll_search", "Failed to set state/county", Error::ERROR, __LINE__);
        return false;
    }

    baseNAD83 = imap::base_83(mapBase->getStateCode(), mapBase->getCountyCode());
    if (coord_type == DECIMAL_LL_SEARCH)
    {
        sscanf(get_value("LAT").c_str(), "%lf", &tmpLat);
        sscanf(get_value("LON").c_str(), "%lf", &tmpLon);
        report("Map_Connection::ll_search", "DEC Lat %lf Lon %lf", Error::DEBUG, __LINE__, tmpLat, tmpLon);
    }
    else if (coord_type == DMS_LL_SEARCH)
    {
        double latDeg, lonDeg, latMin, lonMin, latSec, lonSec;
        sscanf(get_value("LAT_DEG").c_str(), "%lf", &latDeg);
        sscanf(get_value("LAT_MIN").c_str(), "%lf", &latMin);
        sscanf(get_value("LAT_SEC").c_str(), "%lf", &latSec);
        sscanf(get_value("LON_DEG").c_str(), "%lf", &lonDeg);
        sscanf(get_value("LON_MIN").c_str(), "%lf", &lonMin);
        sscanf(get_value("LON_SEC").c_str(), "%lf", &lonSec);
        report("Map_Connection::ll_search", "DMS Lat %lf %lf %lf Lon %lf %lf %lf", Error::DEBUG, __LINE__, latDeg,latMin, latSec, lonDeg,lonMin, lonSec);
        tmpLat = NADConvt::cvtToDecimal(latDeg, latMin, latSec);
        tmpLon = NADConvt::cvtToDecimal(lonDeg, lonMin, lonSec);
    }
    else if (coord_type == SPCS_LL_SEARCH)
    {
        double northing, easting;
        int zone;
        sscanf(get_value("NORTHING").c_str(), "%lf", &northing);
        sscanf(get_value("EASTING").c_str(), "%lf", &easting);
        zone = atoi(get_value("ZONE").c_str());
        report("Map_Connection::ll_search", "SPCS Northing %lf Easting %lf Zone %d", Error::DEBUG, __LINE__, northing, easting, zone);
        if (requestNAD83)
            SPCConvt::SPCS83toLL(zone, northing, easting, tmpLat, tmpLon);
        else
            SPCConvt::SPCS27toLL(zone, northing, easting, tmpLat, tmpLon);
    }
    else if (coord_type == UTM_LL_SEARCH)
    {
        double northing, easting;
        int zone;
        sscanf(get_value("NORTHING").c_str(), "%lf", &northing);
        sscanf(get_value("EASTING").c_str(), "%lf", &easting);
        zone = atoi(get_value("ZONE").c_str());
        report("Map_Connection::ll_search", "UTM Northing %lf Easting %lf Zone %d", Error::DEBUG, __LINE__, northing, easting, zone);
        if (requestNAD83)
            UTMConvt::UTM83toLL(northing, easting, zone, tmpLat, tmpLon);
        else
        {
            easting = easting / (39.37/12.0);
            northing = northing / (39.37/12.0);
            UTMConvt::UTM27toLL(northing, easting, zone, tmpLat, tmpLon);
        }
    }
    else
    {
        report("Map_Connection::ll_search", "Unknown Coord type %s (DEC %s DMS %s SPCS %s UTM %s)", Error::ERROR, __LINE__, coord_type.c_str(), DECIMAL_LL_SEARCH, DMS_LL_SEARCH, SPCS_LL_SEARCH, UTM_LL_SEARCH);
        return false;
    }

    // Do we need to convert nad type?
    if (baseNAD83 != requestNAD83)
    {
        report("Map_Connection::ll_search", "Converting NAD 27 to/from 83 (base NAD83 %d request NAD83 %d)", Error::DEBUG, __LINE__, baseNAD83, requestNAD83);
        if (baseNAD83)
            NADConvt::convertNad27toNad83(tmpLat, tmpLon, lat, lon);
        else
            NADConvt::convertNad83toNad27(tmpLat, tmpLon, lat, lon);
    }
    else
    {
        lat = tmpLat;
        lon = tmpLon;
    }

    report("Map_Connection::ll_search", "Searching lon %lf Lat %lf", Error::DEBUG, __LINE__, lon, lat);
    try
    {
        Vertex v(lon, lat);
        RectHull county_hull;

        // get county hull
        if ( OvlUtil::getHull( static_cast<short>(mapBase->getStateCode()), static_cast<short>(mapBase->getCountyCode()), county_hull))
        {
            // verify the lat/lon is contained in the county hull
            if (county_hull.contains(v, Vertex::EpsilonGEO))
            {
                // create a locate polygon around the lat/lon, use
                // that rect hull to get the feature list and create
                // an image to return to client
                Locate locate(mapBase->getStateCode(), mapBase->getCountyCode());
                locate.initLocate(v);
                const RectHull & rh = _hull_manager.reset(locate.getShape()->getRectHull());
                create_image(rh);
                if (initialSearch)
                    failedPrimarySearch = false;
                return true;
            }
            else
            {
                report("Map_Connection::ll_search", "Coordinates Lon %lf Lat %lf not in county", Error::INFORMATION, __LINE__, lon, lat);
            }
        }
        else
        {
            report("Map_Connection::ll_search", "Failed to get county hull", Error::WARNING, __LINE__);
        }
    }
    catch (const char * err_msg)
    {
        errRptI(( const_cast< char * >( err_msg )));
    }

    return false;
}


void Map_Connection::test_dump( const std::vector< std::pair< std::string, std::string > > & data )
{
    co << HTTPHTMLHeader() << HTMLDoctype( HTMLDoctype::eTransitional ) << std::endl;
    co << html() << head() << std::endl;
    co << title() << "One Call Concepts, Inc. Test Results" << title() << std::endl;
    co << head() << std::endl;
    co << body() << std::endl;
    co << table() << tbody() << std::endl;
    
    for ( std::vector< std::pair< std::string, std::string > >::const_iterator i = data.begin(); i != data.end(); i++ ){
        co << tr().add( td( i->first )).add( td( i->second )) << std::endl;
    }

    co << tbody() << table() << std::endl;
    co << body() << std::endl;
    co << html() << std::endl;

    co.flush();
}


int Map_Connection::street_search(bool initialSearch)
{
    FeatureList::iterator fiter;

    if (initialSearch)
        failedPrimarySearch = true;

    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights
    std::string state_name = get_value( "state" );
    std::string county_name = get_value( "county" );
    std::string place_name = get_value( "place" );
    std::string street_addr = get_value( "address" );
    std::string street_name = get_value( "street" );
    std::string intersection_name = get_value( "intersection" );

    sutil::uppercase( state_name );
    sutil::uppercase( county_name );
    sutil::uppercase( place_name );
    sutil::uppercase( street_addr );
    sutil::uppercase( street_name );
    sutil::uppercase( intersection_name );
    
    report("Map_Connection::street_search", "state_name:        %s", Error::DEBUG, __LINE__, state_name.c_str());
    report("Map_Connection::street_search", "county_name:       %s", Error::DEBUG, __LINE__, county_name.c_str());
    report("Map_Connection::street_search", "place_name:        %s", Error::DEBUG, __LINE__, place_name.c_str());
    report("Map_Connection::street_search", "street_addr:       %s", Error::DEBUG, __LINE__, street_addr.c_str());
    report("Map_Connection::street_search", "street_name:       %s", Error::DEBUG, __LINE__, street_name.c_str());
    report("Map_Connection::street_search", "intersection_name: %s", Error::DEBUG, __LINE__, intersection_name.c_str());

    AddressQueryInfo main_street;
    main_street.state = state_name.c_str();
    main_street.county = county_name.c_str();
    main_street.place = place_name.c_str();

    OvlQuery primary_query;
    OvlQuery secondary_query;

    if ( get_group_codes( main_street, primary_query ) == false ){
        report("Map_Connection::street_search", "no state/county codes found", Error::WARNING, __LINE__);
        return 0;
    }
    
    AddressQueryInfo intersection_street = main_street;
    
    main_street.aFrom = strip_street_number( street_addr ).c_str();
//    StreetUtil::four_part_streets( false );
    CompositeStreet composite_street;
    StreetUtil::parse_street( street_name.c_str(), &composite_street );

    report("Map_Connection::street_search", "addr: %s", Error::DEBUG, __LINE__, main_street.aFrom.c_str());
    report("Map_Connection::street_search", "main street p: %s n: %s t: %s s: %s", Error::DEBUG, __LINE__, 
                                           composite_street.prefix,
                                           composite_street.name,
                                           composite_street.type,
                                           composite_street.suffix);

    set_street( main_street, composite_street );

    StreetUtil::parse_street( intersection_name.c_str(), &composite_street );

    report("Map_Connection::street_search", "intersection p: %s n: %s t: %s s: %s", Error::DEBUG, __LINE__,
                                            composite_street.prefix,
                                            composite_street.name,
                                            composite_street.type,
                                            composite_street.suffix);
        
    set_street( intersection_street, composite_street );

    primary_query.requestExtended();
    primary_query.requestCombine();
    primary_query.clearType();
    const std::vector<int> searchTypes = config->getSearchTypes();

    for (unsigned int i = 0 ; i < searchTypes.size(); i++)
        primary_query.requestType(static_cast<FeatureDef::Type>(searchTypes[i]));
    if (FeatureDef::getSearch(FeatureDef::Grid) == true)
        primary_query.requestType( FeatureDef::Grid );

    get_place_ids( main_street.place, primary_query );

    if ( _current_feature_list.size() > 0 ){
        _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
    }

    if ( _current_intersection_list.size() > 0 ){
        for (IntersectionPtrList::iterator iter = _current_intersection_list.begin(); iter != _current_intersection_list.end(); iter++)
            delete *iter;
        _current_intersection_list.erase(_current_intersection_list.begin(), _current_intersection_list.end());
    }

    // if the street address is filled in, attempt an address search

    secondary_query = primary_query;
    
    if ( !street_addr.empty()){

        // try a primary search on the main street and house number
        
        if ( primary_search( main_street, primary_query, _current_feature_list )){
            if (initialSearch)
                failedPrimarySearch = false;

            if ( _current_feature_list.size() > 1 ){
                return_feature_list( _current_feature_list );
            }
            else {
                const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                    _highlight_features.push_back(*fiter);
                create_image(rh);
            }
            
            return _current_feature_list.size();
        }

        // if the primary search fails, and there is not a
        // cross-street, then attempt a primary search on the main
        // street, and then a secondary search on the main street.
        
        else if ( intersection_name.empty()){
            secondary_query = primary_query;
            secondary_query.clearAddress();

            if ( secondary_search( main_street, secondary_query, _current_feature_list )){
                if ( _current_feature_list.size() > 1 ){
                    return_feature_list( _current_feature_list );
                }
                else {
                    const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                    for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                        _highlight_features.push_back(*fiter);
                    create_image(rh);
                }
            }
            return _current_feature_list.size();
        }
    }

    if ( _current_feature_list.size() > 0 ){
        _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
    }

    if ( ! intersection_name.empty()){ // this should always be true at this point

        secondary_query.clearAddress();
        primary_query = secondary_query;
        OvlQuery intersection_query = primary_query;

        FeatureList primary_feature_list;
        FeatureList secondary_feature_list;
        FeatureList primary_cross_street_feature_list;
        FeatureList secondary_cross_street_feature_list;
        
        if ( ! primary_search( main_street, primary_query, primary_feature_list )){
            if ( ! secondary_search( main_street, secondary_query, secondary_feature_list )){

                // no primary/secondary match on the main street,
                // return the results of a primary/secondary search on
                // the cross street.

                if ( primary_search( intersection_street, intersection_query, primary_cross_street_feature_list ) ||
                     secondary_search( intersection_street, intersection_query, secondary_cross_street_feature_list )){

                     if (primary_cross_street_feature_list.size() > 0)
                        for (fiter = primary_cross_street_feature_list.begin(); fiter != primary_cross_street_feature_list.end(); fiter++)
                            _current_feature_list.push_back(*fiter);
                     else
                        for (fiter = secondary_cross_street_feature_list.begin(); fiter != secondary_cross_street_feature_list.end(); fiter++)
                            _current_feature_list.push_back(*fiter);

                    if ( _current_feature_list.size() > 1 ){
                        return_feature_list( _current_feature_list );
                    }
                    else {
                        const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                        _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                        for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                            _highlight_features.push_back(*fiter);
                        create_image(rh);
                    }
                }

                return _current_feature_list.size();
            }
            else {

                // secondary search on the main street has succeeded.
        
                if ( primary_search( intersection_street, intersection_query, primary_cross_street_feature_list )){

                    get_intersections( secondary_query, intersection_query, _current_intersection_list );

                    if ( _current_intersection_list.size() == 0 ){

                        // at this point there is not an intersection
                        // of the secondary main street and primary
                        // cross street, so check for an intersection
                        // of the secondary main street and the
                        // secondary cross street
                
                        if ( secondary_search( intersection_street, intersection_query, secondary_cross_street_feature_list )){
                            get_intersections( secondary_query, intersection_query, _current_intersection_list );
                        }
                    }

                    if ( _current_intersection_list.size() == 0 ){

                        // no intersection between main and cross street,
                        // return result of the primary/secondary search on
                        // the main street and cross street.
                    
                        if (primary_feature_list.size() > 0)
                            for (fiter = primary_feature_list.begin(); fiter != primary_feature_list.end(); fiter++)
                                _current_feature_list.push_back(*fiter);
                        else
                            for (fiter = secondary_feature_list.begin(); fiter != secondary_feature_list.end(); fiter++)
                                _current_feature_list.push_back(*fiter);
                        if (primary_cross_street_feature_list.size() > 0)
                            for (fiter = primary_cross_street_feature_list.begin(); fiter != primary_cross_street_feature_list.end(); fiter++)
                                _current_feature_list.push_back(*fiter);
                        else
                            for (fiter = secondary_cross_street_feature_list.begin(); fiter != secondary_cross_street_feature_list.end(); fiter++)
                                _current_feature_list.push_back(*fiter);

                        if ( _current_feature_list.size() > 1 ){
                            return_feature_list( _current_feature_list );
                        }
                        else {
                            const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                            _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                            for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                                _highlight_features.push_back(*fiter);
                            create_image(rh);
                        }
                    }
                    else {

                        // intersection search succeeded

                        _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
                        get_current_feature_list( _current_intersection_list );
                
                        if ( _current_intersection_list.size() > 1 ){
                            return_feature_list( _current_feature_list );
                        }
                        else {
                            Locate locate(mapBase->getStateCode(), mapBase->getCountyCode());
                            locate.initLocate(*(_current_intersection_list.begin()));
                            const RectHull & rh = _hull_manager.reset(locate.getShape()->getRectHull());
                            create_image(rh);
                        }
                    }
                }
                else if ( secondary_search( intersection_street, intersection_query, secondary_cross_street_feature_list )){

                    get_intersections( secondary_query, intersection_query, _current_intersection_list );

                    if ( _current_intersection_list.size() == 0 ){

                        // no intersection between main and cross street,
                        // return result of the primary/secondary search on
                        // the main street and cross street.
                    
                        for (fiter = secondary_feature_list.begin(); fiter != secondary_feature_list.end(); fiter++)
                            _current_feature_list.push_back(*fiter);
                        for (fiter = secondary_cross_street_feature_list.begin(); fiter != secondary_cross_street_feature_list.end(); fiter++)
                            _current_feature_list.push_back(*fiter);

                        if ( _current_feature_list.size() > 1 ){
                            return_feature_list( _current_feature_list );
                        }
                        else {
                            const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                            _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                            for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                                _highlight_features.push_back(*fiter);
                            create_image(rh);
                        }
                    }
                    else {

                        // intersection search succeeded

                        _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
                        get_current_feature_list( _current_intersection_list );
                
                        if ( _current_intersection_list.size() > 1 ){
                            return_feature_list( _current_feature_list );
                        }
                        else {
                            Locate locate(mapBase->getStateCode(), mapBase->getCountyCode());
                            locate.initLocate(*(_current_intersection_list.begin()));
                            const RectHull & rh = _hull_manager.reset(locate.getShape()->getRectHull());
                            create_image(rh);
                        }
                    }
                }
                else {

                    // no primary/secondary match found on the cross street 
                    // return result of the secondary search on
                    // the main street
                    
                    for (fiter = secondary_feature_list.begin(); fiter != secondary_feature_list.end(); fiter++)
                        _current_feature_list.push_back(*fiter);
                    
                    if ( _current_feature_list.size() > 1 ){
                        return_feature_list( _current_feature_list );
                    }
                    else {
                        const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                        _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                        for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                            _highlight_features.push_back(*fiter);
                        create_image(rh);
                    }
                }
            }
        }
        else {
            
            // primary search on main street has succeeded

            if ( primary_search( intersection_street, intersection_query, primary_cross_street_feature_list )){

                // primary search on cross street has succeeded


                get_intersections( primary_query, intersection_query, _current_intersection_list );
                if (_current_intersection_list.size() > 0 && initialSearch)
                    failedPrimarySearch = false;


                if ( _current_intersection_list.size() == 0 ){

                    // at this point there is not an intersection
                    // of the primary main street and primary
                    // cross street, so check for an intersection
                    // of the secondary main street and the
                    // primary cross street
                
                    secondary_search( main_street, secondary_query, secondary_feature_list );
                    get_intersections( secondary_query, intersection_query, _current_intersection_list );
                }
                
                if ( _current_intersection_list.size() == 0 ){

                    // at this point there is not an intersection
                    // of the primary or secondary main street and primary
                    // cross street, so check for an intersection
                    // of the primary main street and the
                    // secondary cross street
                
                    secondary_search( intersection_street, intersection_query, secondary_cross_street_feature_list );
                    get_intersections( primary_query, intersection_query, _current_intersection_list );
                }

                if ( _current_intersection_list.size() == 0 ){

                    // at this point there is not an intersection of
                    // the primary or secondary main street and
                    // primary cross street, or the primary main
                    // street with the secondary cross street so check
                    // for an intersection of the secondary main
                    // street and the secondary cross street
                
                    get_intersections( secondary_query, intersection_query, _current_intersection_list );
                }

                if ( _current_intersection_list.size() == 0 ){

                    // no intersection between main and cross street,
                    // return result of the primary/secondary search on
                    // the main street and cross street.
                    
                    for (fiter = primary_feature_list.begin(); fiter != primary_feature_list.end(); fiter++)
                        _current_feature_list.push_back(*fiter);
                    if (primary_cross_street_feature_list.size() > 0)
                        for (fiter = primary_cross_street_feature_list.begin(); fiter != primary_cross_street_feature_list.end(); fiter++)
                            _current_feature_list.push_back(*fiter);
                    else
                        for (fiter = secondary_cross_street_feature_list.begin(); fiter != secondary_cross_street_feature_list.end(); fiter++)
                            _current_feature_list.push_back(*fiter);

                    if ( _current_feature_list.size() > 1 ){
                        return_feature_list( _current_feature_list );
                    }
                    else {
                        const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                        _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                        for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                            _highlight_features.push_back(*fiter);
                        create_image(rh);
                    }
                }
                else {

                    // intersection search succeeded

                    _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
                    get_current_feature_list( _current_intersection_list );

                    if ( _current_intersection_list.size() > 1 ){
                        return_feature_list( _current_feature_list );
                    }
                    else {
                        Locate locate(mapBase->getStateCode(), mapBase->getCountyCode());
                        locate.initLocate(*(_current_intersection_list.begin()));
                        const RectHull & rh = _hull_manager.reset(locate.getShape()->getRectHull());
                        create_image(rh);
                    }
                }
            }
            else if ( secondary_search( intersection_street, intersection_query, secondary_cross_street_feature_list )){

                // secondary search on cross street has succeeded

                get_intersections( primary_query, intersection_query, _current_intersection_list );

                if ( _current_intersection_list.size() == 0 ){

                    // at this point there is not an intersection
                    // of the primary main street and primary/secondary
                    // cross street, so check for an intersection
                    // of the secondary main street and the
                    // secondary cross street
                
                    secondary_search( main_street, secondary_query, secondary_feature_list );
                    get_intersections( secondary_query, intersection_query, _current_intersection_list );
                }
                
                if ( _current_intersection_list.size() == 0 ){
                    
                    // no intersection between main and cross street,
                    // return result of the primary/secondary search on
                    // the main street and cross street.
                    
                    for (fiter = primary_feature_list.begin(); fiter != primary_feature_list.end(); fiter++)
                        _current_feature_list.push_back(*fiter);
                    for (fiter = secondary_cross_street_feature_list.begin(); fiter != secondary_cross_street_feature_list.end(); fiter++)
                        _current_feature_list.push_back(*fiter);

                    if ( _current_feature_list.size() > 1 ){
                        return_feature_list( _current_feature_list );
                    }
                    else {
                        const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                        _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                        for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                            _highlight_features.push_back(*fiter);
                        create_image(rh);
                    }
                }
                else {

                    // intersection search succeeded

                    _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
                    get_current_feature_list( _current_intersection_list );

                    if ( _current_intersection_list.size() > 1 ){
                        return_feature_list( _current_feature_list );
                    }
                    else {
                        Locate locate(mapBase->getStateCode(), mapBase->getCountyCode());
                        locate.initLocate(*(_current_intersection_list.begin()));
                        const RectHull & rh = _hull_manager.reset(locate.getShape()->getRectHull());
                        create_image(rh);
                    }
                }
            }
            else {

                // no primary/secondary match found on the cross street 
                // return result of the secondary search on
                // the main street
                    
                for (fiter = primary_feature_list.begin(); fiter != primary_feature_list.end(); fiter++)
                    _current_feature_list.push_back(*fiter);
                    
                if ( _current_feature_list.size() > 1 ){
                    return_feature_list( _current_feature_list );
                }
                else {
                    const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );
                    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
                    for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
                        _highlight_features.push_back(*fiter);
                    create_image(rh);
                }
            }
        }
    }

    return _current_feature_list.size();
}


void Map_Connection::return_feature_list( const FeatureList & features )
{
    co << "Content-Type: application/octet-stream\n\n";

    std::ostringstream msg;
    
    msg << "FEATURLIST," << _id;

    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());

    std::string tmp( msg.str());
    co << header << tmp;
    
    std::string fname;
    std::string addr;
    
    for (FeatureList::const_iterator iter = features.begin(); iter != features.end(); iter++)
    {
        mapBase->getNameManager().getFullFeatureNameString((*iter).getMainName(), fname);

        if ( fname.length() )
        {
            co << "~";
            
            if ((*iter).hasExtended() && (*iter).hasAddresses())
            {
                addr = L_BEG;
                get_addr_range((*iter).getAddressesLeft(), 6, addr);
                addr += R_BEG;
                get_addr_range((*iter).getAddressesRight(), 6, addr);
                co << addr << " ";

                report("Map_Connection::return_feature_list", "addr : %s", Error::DEBUG, __LINE__, addr.c_str());
            }
            else
            {
                co << "                                ";

                report("Map_Connection::return_feature_list", "addr : NULL", Error::DEBUG, __LINE__);
            }

            justLeft(fname,  25 ); fname = padRight(fname, 25 ); fname.erase(25);

            if ( addr.length() == 31 ){
                co << "                          ";
            }
                
            co << fname;

            std::string place = mapBase->getNameManager().getPlaceString((*iter).getMcdRight());
            justLeft(place, 25); place = padRight(place, 25); place.erase(25);

            co << " " << place;
            
            report("Map_Connection::return_feature_list", "fname: %s", Error::DEBUG, __LINE__, fname.c_str());
            report("Map_Connection::return_feature_list", "place: %s", Error::DEBUG, __LINE__, place.c_str());
        }
    }

    co.flush();
}


int Map_Connection::dump_features(const FeatureList & features)
{
    std::string fname;
    
    for (FeatureList::const_iterator iter = features.begin(); iter != features.end(); iter++)
    {
        mapBase->getNameManager().getFullFeatureNameString((*iter).getMainName(), fname);
        if (fname.length())
            report("Map_Connection::dump_features", fname.c_str(), Error::DEBUG, __LINE__);
    }

    return features.size();
}


void Map_Connection::transmit_image(bool includeHeader)
{
    if (!includeHeader || _hull_manager.get_initialized())
    {
        // send the image

        report("Map_Connection::transmit_image", "send the message", Error::DEBUG, __LINE__);

        co << "Content-Type: application/octet-stream\n\n";

        if (includeHeader)
        {
            std::ostringstream msg;

            msg.precision( 8 );
    
            if ( send_shutdown_ ){
                send_shutdown_ = false;
                msg << shutdown_message_;
            }

            msg << "IMAGE," << _id << "," << _x_pixels << "," << _y_pixels << "," << _scale << ","
                << _hull_manager.get_current_hull()[NW].x() << ","
                << _hull_manager.get_current_hull()[NW].y() << ","
                << _hull_manager.get_current_hull()[SE].x() << ","
                << _hull_manager.get_current_hull()[SE].y() << std::ends;

            char header[ 11 ];
            sprintf( header, "%-10d", msg.str().size());

            report("Map_Connection::transmit_image", "message header: %s%s", Error::DEBUG, __LINE__, header, msg.str().c_str());
brianX=0;
brianY=0;
            std::string tmp( msg.str());
            co << header << tmp;
        }

        report("Map_Connection::transmit_image", "send the message data image_ptr = %p, image_size = %d", Error::DEBUG, __LINE__, _image_ptr, _image_size);

        co.write( (const char *)_image_ptr, _image_size );
        co.flush();
    

        if ( co.bad())
            report("Map_Connection::transmit_image", "error sending image", Error::WARNING, __LINE__);

        gdFree( _image_ptr );

        _image_ptr = NULL;
        _image_size = 0;
    }
    else
    {
        report("Map_Connection::transmit_image", "hull manager not initialized", Error::ERROR, __LINE__);
        draw_error_page("hull manager not initialized");
    }

    return;
}


void Map_Connection::draw_exit_page()
{
    std::string content = "0;url=" + config->login_url();
    report("Map_Connection::draw_exit_page", "content: %s", Error::DEBUG, __LINE__, content.c_str());

    co << HTTPHTMLHeader() << HTMLDoctype( HTMLDoctype::eTransitional ) << std::endl;
    co << html() << head() << std::endl;
    co << meta().set( "http-equiv", "refresh" ).set( "content", content ) << std::endl;
    co << title() << "One Call Concepts, Inc." << title() << std::endl;
    co << head() << std::endl;
    co << body() << 
        "Returning to login page" << std::endl;
    co << body() << std::endl;
    co << html() << std::endl;

    co.flush();
}


void Map_Connection::unknown_request( const std::string & op_code )
{
    report("Map_Connection::unknown_request", "Unknown op_code \"%s\", sending back error", Error::ERROR, __LINE__, op_code.c_str());

    co << HTTPHTMLHeader() <<
        html()<<
        head() <<
        "<meta http-equiv=\"refresh\" content=redirect>" <<
        head() <<
        body() << 
        "The " << b() << op_code << b() << " has not been implemented." <<
        cgicc::form() << "<input type=\"hidden\" name=\"op_code\" value=\"back\">" << cgicc::form() <<
        body() << 
        html() <<
        std::endl;

    co.flush();
}


bool Map_Connection::create_locate_poly_string(const Polygon& poly, std::ostringstream& poly_buf)
{
    if (poly.size() > 0)
    {
        int pos = 0;

        poly_buf << poly.size() + 1 << " " << std::setprecision( 10 ) << poly[pos].x() << " " << poly[pos].y() << " ";
        pos++;
        while (pos < poly.size())
        {
            poly_buf << poly[pos].x() << " " << poly[pos].y() << " ";
            pos++;
        }

        pos = 0;
        poly_buf << poly[pos].x() << " " << poly[pos].y();

        errRptD(( "poly_buf: %s", poly_buf.str().c_str() ));
        return true;
    }
    return false;
}
// ****************************************************************************
// NAME:        Map_Connection::transmit_locate_polygon
//
// DESCRIPTION: this method returns the locate polygon to the imap client
//
// PARAMETERS:  none
//
// RETURNS:     void
//
// ****************************************************************************
void Map_Connection::transmit_locate_polygon()
{
    if (_hull_manager.get_initialized())
    {
        Polygon & locate_poly = _hull_manager.get_locate_polygon();
        std::ostringstream poly_buf;

        if (create_locate_poly_string(locate_poly, poly_buf))
        {
            co << "Content-Type: application/octet-stream\n\n";

            std::ostringstream msg;

            msg << "LOCATEPOLY,";

            if ( _hull_manager.get_polygon_is_grid())
                msg << "Y" << std::ends;
            else
                msg << "N" << std::ends;

            char header[ 11 ];
            sprintf( header, "%-10d", msg.str().size());
            string tmp( msg.str());

            co << header << tmp;

            co << poly_buf.str();
            co.flush();
            return;
        }
        else {
            errRptW(( "no locate polygon found" ));
            draw_error_page( "no locate polygon found" );
        }
    }
    else {
        report("Map_Connection::transmit_locate_polygon", "hull manager not initialized", Error::ERROR, __LINE__);
        draw_error_page("hull manager not initialized");
    }
}


bool Map_Connection::get_scaled_image()
{
    Vertex nw;
    Vertex se;
    double dx, dy, scale_pct;

    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights
            
    if (!imap::setStateCounty(get_value("STATE"), get_value("COUNTY"), mapBase))
    {
        report("Map_Connection::get_scaled_image", "Failed to set state/county", Error::ERROR, __LINE__);
        return false;
    }

    sscanf( get_value( "NWX" ).c_str(), "%lf", &dx );
    sscanf( get_value( "NWY" ).c_str(), "%lf", &dy );

    nw.set( dx, dy );
    
    sscanf( get_value( "SEX" ).c_str(), "%lf", &dx );
    sscanf( get_value( "SEY" ).c_str(), "%lf", &dy );

    se.set( dx, dy );

    sscanf( get_value( "ZOOM-PCT" ).c_str(), "%lf", &scale_pct );

    errRptI(( "scale hull (%f,%f)-(%f,%f) by %f", nw.x(), nw.y(), se.x(), se.y(), scale_pct ));
    
    RectHull rh( nw, se );
    rh *= ( 100.0 + scale_pct ) / 100.0;

    // Determine if the max/min zoom reached
    double newScale = rh.diagonalGEO();
    if (newScale == 0.0 || newScale < mapBase->getMinimumScale() || newScale > mapBase->getMaximumScale())
    {
        errRptI(( "max scaling reached %f < %f < %f", mapBase->getMinimumScale(), newScale, mapBase->getMaximumScale()));
        return false;
    }
    else
    {
        errRptI(( "new hull: (%f,%f)-(%f,%f)", rh[NW].x(), rh[NW].y(), rh[SE].x(), rh[SE].y()));

        _hull_manager.reset_no_minimum( rh );

        create_image(rh);
        return true;
    }
}

void Map_Connection::get_identify_features()
{
    OvlQuery pickQuery;
    double dx, dy;
    double pointX, pointY;
    Vertex pickPt;
    Vertex nw, se;
    FeatureList::iterator fiter;
    DrawSetupInfo info;

    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights

    if (!imap::setStateCounty(get_value( "STATE" ), get_value( "COUNTY" ), mapBase))
    {
        report("Map_Connection::get_identify_features", "Failed to set state/county", Error::ERROR, __LINE__);
        return;
    }

    sscanf( get_value( "POINTX" ).c_str(), "%lf", &pointX );
    sscanf( get_value( "POINTY" ).c_str(), "%lf", &pointY );
    pickPt.set(pointX, pointY);
    errRptD(( "point (%f,%f)", pickPt.x(), pickPt.y()));

    sscanf( get_value( "NWX" ).c_str(), "%lf", &dx );
    sscanf( get_value( "NWY" ).c_str(), "%lf", &dy );
    nw.set( dx, dy );

    sscanf( get_value( "SEX" ).c_str(), "%lf", &dx );
    sscanf( get_value( "SEY" ).c_str(), "%lf", &dy );
    se.set( dx, dy );
    errRptI(( "request hull (%f,%f)-(%f,%f)", nw.x(), nw.y(), se.x(), se.y()));
    RectHull hull( nw, se );

    // Need to calculate zoom level
    setupImageInfo(info, hull);

    // Build pick query
    pickQuery.requestState(mapBase->getStateCode());
    pickQuery.requestCounty(mapBase->getCountyCode());
    pickQuery.requestExtended();
    pickQuery.requestDisplayLevel(PageDraw::calculateDisplayLevel(info.scaleMagnification));
    pickQuery.requestGeneralize();
    pickQuery.requestCombine();
    pickQuery.requestHull(hull);
    FeatureDef::Type  ft;
    FeatureTypeLoop(ft)
        if (FeatureDef::getDisplay(ft) == true)
            pickQuery.requestType(ft);

    // Get the features
    FeatureList tfList, copies;
    if (OvlMgr::getFeatures(pickQuery, tfList) == false)
        return;
    // Get rid of copied features
    if (OvlUtil::removeCopies(tfList, copies))
        copies.erase(copies.begin(), copies.end()) ;

    // Pick radius is a percentage of scale
    double scale = calculate_scale();
    long distForRound = (long) (scale * Vertex::FEET_PER_MILE) ;
    double baseRadius = config->get_identify_percent_radius() * distForRound;
    // We need a pick radius of at least 5 feet
    if (baseRadius < 5)
        baseRadius = 5;
    double pickRadius = baseRadius / Vertex::FEET_PER_MILE / Vertex::MILES_PER_DEGREE;

    Vertex tmpPt(pickPt.x(), pickPt.y() + pickRadius);
    double radMiles = pickPt.distanceGEO(tmpPt);

    for (fiter = tfList.begin(); fiter != tfList.end(); fiter++)
    {
        const Geometric& geom = (*fiter).getGeometric();
        const Shape* shape = geom.getShape();
        if (shape == NULL)
            continue;
        if (shape->distanceGEO(pickPt) <= radMiles)
            _highlight_features.push_back(*fiter);
    }

    co << "Content-Type: application/octet-stream\n\n";

    std::ostringstream msg;
    
    msg << "IDENTLIST," << _id << std::ends;

    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());

    string tmp( msg.str());
    co << header << tmp;

    int featuresReturned = 0;
    int count = 0;
    std::string fname;
    std::string addr;
    std::string place;

    // Since the message may or may not have aliases, adding an extra ~ to end helps client parse the string

    if (config->get_number_identify_features() > _highlight_features.size())
        featuresReturned = _highlight_features.size();
    else
        featuresReturned = config->get_number_identify_features();

    co << "~" << featuresReturned << "~" << _highlight_features.size() << "~";
    if (_highlight_features.size() == 0)
    {
        // No highlights, we are done with the message
        co.flush();
        return;
    }

    for (fiter = _highlight_features.begin(); fiter != _highlight_features.end(); fiter++)
    {
        if (count >= featuresReturned)
            break;
        count++;
        fname = buildFullNameString(mapBase->getNameManager(), (*fiter).getMainName());
        co << (*fiter).getRecID() << "~" << fname << "~";
        report("Map_Connection::get_identify_features", "fname: %s", Error::DEBUG, __LINE__, fname.c_str());

        build_identify_address_string(*fiter, addr);
        report("Map_Connection::get_identify_features", "addr: %s", Error::DEBUG, __LINE__, addr.c_str());
        co << addr << "~";

        place = mapBase->getNameManager().getPlaceString( (*fiter).getMcdLeft());
        report("Map_Connection::get_identify_features", "place left: %s", Error::DEBUG, __LINE__, place.c_str());
        co << place << "~";
        place = mapBase->getNameManager().getPlaceString( (*fiter).getMcdRight());
        report("Map_Connection::get_identify_features", "place right: %s", Error::DEBUG, __LINE__, place.c_str());
        co << place << "~";

        if ( (*fiter).hasNames())
        {
            const FNameList   &nameList = (*fiter).getNames() ;
            std::string text;

            if ((*fiter).getType() == FeatureDef::Boundary ||
                (*fiter).getType() == FeatureDef::CountyBoundary)
            {
                if (nameList.size())
                {
                    text = buildFullPlaceString(mapBase->getNameManager(), *(nameList.begin()));
                    if (! isBlank(text))
                    {
                        report("Map_Connection::get_identify_features", "names: %s", Error::DEBUG, __LINE__, text.c_str());
                        co << "1~" << text << "~";
                    }
                    else
                        co << "0~";
                }
                else
                    co << "0~";
            }
            else
            {
                co << nameList.size() << "~";
                for (FNameList::const_iterator niter = nameList.begin(); niter != nameList.end(); niter++)
                {
                    text = buildFullNameString(mapBase->getNameManager(), *niter);
                    report("Map_Connection::get_identify_features", "names: %s", Error::DEBUG, __LINE__, text.c_str());
                    co << text << "~";
                }
            }
        }
        else
        {
            co << "0~";
        }
    }

    co << std::ends;
    co.flush();

}

void Map_Connection::build_identify_address_string(const Feature& feat, std::string& addr)
{
    addr = "";
    if (feat.hasExtended() && feat.hasAddresses())
    {
        int maxAddress = 0, minAddress = 99999;
        // Find maximum and minimum address
        if (feat.getAddressesLeft().size())
            build_identify_address_string_for_side(feat.getAddressesLeft(), maxAddress, minAddress);
        if (feat.getAddressesRight().size())
            build_identify_address_string_for_side(feat.getAddressesRight(), maxAddress, minAddress);
        std::string fname = buildFullNameString(mapBase->getNameManager(), feat.getMainName());
        addr = ntos((long)minAddress) + " to " + ntos((long)maxAddress) + " " + fname;
    }
}

void Map_Connection::build_identify_address_string_for_side(const FAddressList &faList, int &maxAddress, int &minAddress)
{
    for (FAddressList::const_iterator iter = faList.begin(); iter != faList.end(); iter++)
    {
        if (!((*iter).getFromCompassF()) && !((*iter).getToCompassF()))
        {
            if ((*iter).getFromAddress() > maxAddress)
                maxAddress = (*iter).getFromAddress();
            if ((*iter).getToAddress()   > maxAddress)
                maxAddress = (*iter).getToAddress();
            if ((*iter).getFromAddress() < minAddress)
                minAddress = (*iter).getFromAddress();
            if ((*iter).getToAddress()   < minAddress)
                minAddress = (*iter).getToAddress();
        }
        else
        {
            if ((*iter).getFromCoordNumber() > maxAddress)
                maxAddress = (*iter).getFromCoordNumber();
            if ((*iter).getToCoordNumber()   > maxAddress)
                maxAddress = (*iter).getToCoordNumber();
            if ((*iter).getFromCoordNumber() < minAddress)
                minAddress = (*iter).getFromCoordNumber();
            if ((*iter).getToCoordNumber()   < minAddress)
                minAddress = (*iter).getToCoordNumber();
        }
    }
}

bool Map_Connection::get_identify_image()
{
    Vertex nw;
    Vertex se;
    double dx, dy;

    // Don't erase identifies, we need them here!
    sscanf( get_value( "NWX" ).c_str(), "%lf", &dx );
    sscanf( get_value( "NWY" ).c_str(), "%lf", &dy );
    nw.set( dx, dy );

    sscanf( get_value( "SEX" ).c_str(), "%lf", &dx );
    sscanf( get_value( "SEY" ).c_str(), "%lf", &dy );
    se.set( dx, dy );
    errRptI(( "request hull (%f,%f)-(%f,%f)", nw.x(), nw.y(), se.x(), se.y()));

    RectHull rh( nw, se );
    errRptI(( "new hull: (%f,%f)-(%f,%f)", rh[NW].x(), rh[NW].y(), rh[SE].x(), rh[SE].y()));

    _hull_manager.reset_no_minimum( rh );

    create_image(rh);
    return true;
}

void Map_Connection::get_image_from_page()
{
    if (_hull_manager.get_initialized())
    {
        int x, y, z;

        sscanf( get_value( "XX" ).c_str(), "%d", &x );
        sscanf( get_value( "YY" ).c_str(), "%d", &y );
        sscanf( get_value( "ZZ" ).c_str(), "%d", &z );
brianX = x;
brianY = y;
 
        errRptD(( "get_image_from_page( %d, %d, %d )", x, y, z ));

        RectHull & rh = _hull_manager.new_page( x, y, z );
        create_image(rh);
        transmit_image();
    }
    else
    {
        report("Map_Connection::get_image_from_page", "hull manager not initialized", Error::ERROR, __LINE__);
        draw_error_page("hull manager not initialized");
    }
}


void Map_Connection::return_server_error()
{
    co << "Content-Type: application/octet-stream\n\n";

    std::ostringstream msg;

    msg << "SERVER ERROR" << std::ends;
    char header[ 11 ];

    sprintf( header, "%-10d", msg.str().size());
    std::string tmp( msg.str());
    co << header << tmp;
    
    co.flush();
}


void Map_Connection::return_error_image(bool sendJPG)
{
    co << "Content-Type: application/octet-stream\n\n";

    if (! sendJPG)
    { 
        std::ostringstream msg;
        msg << "SEARCH FAILED," << _id << std::ends;
        char header[ 11 ];
        sprintf( header, "%-10d", msg.str().size());
        std::string tmp( msg.str());
        co << header << tmp;
    }
    else
    {
        SysParm aParm(SysParmStr::ERROR_JPG_IMAGE_PARM);
        std::string errorImage = aParm.getValue(DEFAULT_ERROR_JPG_IMAGE);
        FILE *in;
        in = fopen(errorImage.c_str(), "rb");
        if (in != NULL)
        {
            gdImagePtr image;
            image = gdImageCreateFromJpeg(in);
            if (image != NULL)
            {
                int image_size;
                void* image_ptr = gdImageJpegPtr( image, &image_size, -1 );
                co.write( (const char *)image_ptr, image_size);
                co.flush();
                gdFree(image_ptr);
                gdImageDestroy(image);
            }
            else
            {
            report("Map_Connection::return_error_image", "Failed to create GD image for error image \"%s\"", Error::ERROR, __LINE__, errorImage.c_str());
            }
            fclose(in);
        }
        else
        {
            report("Map_Connection::return_error_image", "Failed to open error image \"%s\"", Error::ERROR, __LINE__, errorImage.c_str());
        }
    }

    co.flush();
    
}

std::pair< int, int > Map_Connection::get_addr_range( const FAddressList & faddr_list, int width, std::string & str )
    // see bulidSideString()...
{
    static const int    ADDR_HUGE_INT = 30000;
    static const char * L_R_SEP           = ":";
    static const char * PRES_STR_ZEROES   =  "0";
    
    int maxAddress = 0, minAddress = ADDR_HUGE_INT;
    int addr_cnt = 0;
    
    for (FAddressList::const_iterator iter = faddr_list.begin(); iter != faddr_list.end(); iter++)
    {

        if ( addr_cnt == 2 ){
            break;
        }

        addr_cnt++;
        
        if ( !(*iter).getFromCompassF() && !(*iter).getToCompassF())
        {
            if ((*iter).getFromAddress() > maxAddress)
                maxAddress = (*iter).getFromAddress();
            if ((*iter).getToAddress()   > maxAddress)
                maxAddress = (*iter).getToAddress();
            if ((*iter).getFromAddress() < minAddress)
                minAddress = (*iter).getFromAddress();
            if ((*iter).getToAddress()   < minAddress)
                minAddress = (*iter).getToAddress();

            report("Map_Connection::get_addr_range", "min: %d", Error::DEBUG, __LINE__, minAddress);
            report("Map_Connection::get_addr_range", "max: %d", Error::DEBUG, __LINE__, maxAddress);

            str += ntos(minAddress, width, PRES_STR_ZEROES) + L_R_SEP + ntos(maxAddress, width, PRES_STR_ZEROES);

            report("Map_Connection::get_addr_range", "str: %s", Error::DEBUG, __LINE__, str.c_str());
        }
        else
        {
            char fdir, tdir, mindir[2], maxdir[2];
            fdir = (*iter).getFromCompass();
            tdir = (*iter).getToCompass();

            mindir[0] = '\0';
            maxdir[0] = '\0';
            
            if ((*iter).getFromCoordNumber() > maxAddress) 
            { 
                maxAddress = (*iter).getFromCoordNumber();
                maxdir[0] = fdir ;
                maxdir[1] = '\0';                
            }
            if ((*iter).getToCoordNumber() > maxAddress) 
            {
                maxAddress = (*iter).getToCoordNumber();
                maxdir[0] = tdir ;              
                maxdir[1] = '\0';                
            }
            if ((*iter).getFromCoordNumber() < minAddress) 
            {
                minAddress = (*iter).getFromCoordNumber();
                mindir[0] = fdir ;
                mindir[1] = '\0';               
            }
            if ((*iter).getToCoordNumber() < minAddress) 
            {
                minAddress = (*iter).getToCoordNumber();
                mindir[0] = tdir ;
                mindir[1] = '\0';                
            }
            
            str += mindir + ntos(minAddress, width, PRES_STR_ZEROES) + L_R_SEP + maxdir + ntos(maxAddress, width, PRES_STR_ZEROES) ;

            report("Map_Connection::get_addr_range", "str: %s", Error::DEBUG, __LINE__, str.c_str());
        }
    }

    return std::make_pair( minAddress, maxAddress );
}


void Map_Connection::get_district_feature(uint index)
{
    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights
    report("Map_Connection::get_district_feature", "feature id: %d", Error::DEBUG, __LINE__, index);
    report("Map_Connection::get_district_feature", "feature list length: %d", Error::DEBUG, __LINE__, _current_feature_list.size());

    int    featId = index;
    bool   recFound = false;
    for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
    {
        if ( featId == fiter->getRecID())
        {
            const RectHull & rh = _hull_manager.reset((*fiter).getHull());
            create_image(rh);
            transmit_image();
            recFound = true;
            break;
        }
    }
    if (! recFound)
    {
        report("Map_Connection::get_district_feature", "invalid feature requested %d", Error::NOTICE, __LINE__, featId);
        return_error_image();
    }
}

bool Map_Connection::build_feature_list_from_parms(const std::string& feature_indices, const std::string& trans_indices, FeatureList &multiFeatureList)
{
    std::vector<std::string> stringList;
    TransID tran_id;

    sutil::split(feature_indices, stringList, ",");
    for ( std::vector<std::string>::const_iterator i = stringList.begin(); i != stringList.end(); i++ )
    {
        int curID = atoi((*i).c_str());
        bool recFound = false;
        for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
        {
            if ( curID == fiter->getRecID())
            {
                multiFeatureList.push_back(*fiter);
                recFound = true;
                break;
            }
        }
        if (! recFound)
        {
            report("Map_Connection::build_feature_list_from_parms", "invalid feature requested %d", Error::NOTICE, __LINE__, curID);
            return false;
        }
    }

    stringList.erase(stringList.begin(), stringList.end());
    sutil::split(trans_indices, stringList, ",");
    for ( std::vector<std::string>::const_iterator i = stringList.begin(); i != stringList.end(); i++ )
    {
        tran_id = atoi((*i).c_str());
        TransRecord rec;
        msl::TransactionDB db(config->getStateCode());
        TransFeatureData tfd;
        RectHull tranHull;
        Geometric geo;
        bool found = false;

        if ( db.getTrans( tran_id, rec ))
        {
            ulong feat_id = rec.getNewFeatId();

            if ( feat_id <= 0 )
            {
                feat_id = rec.getOrigFeatId();
            }

            if ( db.getTransData( feat_id, tfd ))
            {
                Feature feat;
                tfd.getHull(tranHull);
                feat.setHull(tranHull);
                tfd.getGeometric(geo);
                feat.setGeometric(geo);
                multiFeatureList.push_back(feat);
                found = true;
            }
        }
        if (! found)
        {
            report("Map_Connection::build_feature_list_from_parms", "invalid trans requested %d", Error::NOTICE, __LINE__, tran_id);
            return false;
        }
    }
    return true;
}


void Map_Connection::get_multi_district_features(const string& feature_indices, const string& trans_indices)
{
    FeatureList multiFeatureList;

    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights
    report("Map_Connection::get_multi_district_features", "feature indices: %s", Error::DEBUG, __LINE__, feature_indices.c_str());
    report("Map_Connection::get_multi_district_features", "trans indices: %s", Error::DEBUG, __LINE__, trans_indices.c_str());
    report("Map_Connection::get_multi_district_features", "feature list length: %d", Error::DEBUG, __LINE__, _current_feature_list.size());

    if (!build_feature_list_from_parms(feature_indices, trans_indices, multiFeatureList))
    {
            return_error_image();
            return;
    }

    if (multiFeatureList.size() == 0)
    {
        report("Map_Connection::get_multi_district_features", "No feature found to display", Error::NOTICE, __LINE__);
        return_error_image();
        return;
    }

    const RectHull & rh = _hull_manager.reset(multiFeatureList, mapBase->getStateCode(), mapBase->getCountyCode());
    if (!create_image(rh))
    {
        return_error_image();
        return;
    }
    transmit_image();
}

void Map_Connection::get_feature(uint index)
{
    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights
    report("Map_Connection::get_feature", "feature index: %d", Error::DEBUG, __LINE__, index);
    report("Map_Connection::get_feature", "feature list length: %d", Error::DEBUG, __LINE__, _current_feature_list.size());

    if ( index > _current_feature_list.size()){
        report("Map_Connection::get_feature", "invalid feature requested", Error::NOTICE, __LINE__);
        return_error_image();
        return;
    }

    FeatureList::iterator fiter = _current_feature_list.begin();
    uint count = 0;
    for (; fiter != _current_feature_list.end() && count < index; fiter++, count++); // yes, I mean ;

    Locate * locate = NULL;
    locate = new Locate(mapBase->getStateCode(), mapBase->getCountyCode());
    if ( !locate ){
        report("Map_Connection::get_feature", "cannot create locate object from feature", Error::NOTICE, __LINE__);
        return_error_image();
        return;
    }
    
    if ( _current_intersection_list.size() > 0 )
    {
        count = 0;
        IntersectionPtrList::iterator iiter = _current_intersection_list.begin();
        for (; iiter != _current_intersection_list.end() && count < index; iiter++, count++); // yes, I mean ;
        locate->initLocate((*iiter)->point());

        if ( !locate ){
            report("Map_Connection::get_feature", "cannot create locate object from feature", Error::NOTICE, __LINE__);
            delete locate;
            return;
        }
        
        _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
        for (FeatureList::const_iterator cfiter = locate->getFeatures().begin(); cfiter != locate->getFeatures().end(); cfiter++)
            _highlight_features.push_back(*cfiter);

        const RectHull & rh = _hull_manager.reset(locate->getShape()->getRectHull());


        delete locate;

        create_image(rh);
        transmit_image();
    }
    else {
        locate->initLocate(*fiter);

        FeatureList save_list;
        for (fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
             save_list.push_back(*fiter);
        
        _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
        _current_feature_list = locate->getFeatures();

        if ( _current_feature_list.size() == 0 ){
            report("Map_Connection::get_feature", "no features found", Error::NOTICE, __LINE__);
            _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
            for (fiter = save_list.begin(); fiter != save_list.end(); fiter++)
                 _current_feature_list.push_back(*fiter);
            return_error_image();
            return;
        }

        _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());
        for (fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
            _highlight_features.push_back(*fiter);

        report("Map_Connection::get_feature", "found %d features", Error::DEBUG, __LINE__, _current_feature_list.size());
    
        const RectHull & rh = _hull_manager.reset( _current_feature_list, mapBase->getStateCode(), mapBase->getCountyCode() );

        report("Map_Connection::get_feature", "found %d features", Error::DEBUG, __LINE__, _current_feature_list.size());
        
        _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
        for (fiter = save_list.begin(); fiter != save_list.end(); fiter++)
            _current_feature_list.push_back(*fiter);

        create_image(rh);
        transmit_image();
        delete locate;
    }
}

bool Map_Connection::get_coords_for_feature(const Feature& feat, std::ostringstream& coords)
{
    bool found = false;
    coords.precision( 8 );
    const Geometric& g = feat.getGeometric();

    if ( g.type() == MSLenum::POLYGON )
    {
        found = true;
        int points = g.size();
        report("Map_Connection::get_coords_for_feature", "points = %d", Error::DEBUG, __LINE__, points);

        const Polygon * p = static_cast< const Polygon * >( g.getShape() );

        coords << p->size();
    
        for (int i = 0; i < p->size(); i++)
        {
            coords << " " << (*p)[i].x() << " " << (*p)[i].y();
        }
    }

    report("Map_Connection::get_coords_for_feature", "snapshot", Error::DEBUG, __LINE__);
    return found;
}

void Map_Connection::get_feature_coords( uint index )
{
    report("Map_Connection::get_feature_coords", "feature featId: %d", Error::DEBUG, __LINE__, index);
    report("Map_Connection::get_feature_coords", "feature list length: %d", Error::DEBUG, __LINE__, _current_feature_list.size());

    int   featId = index;
    bool  recFound = false;
    FeatureList::iterator fiter = _current_feature_list.begin();
    for (; fiter != _current_feature_list.end(); fiter++)
    {
        if ( featId == fiter->getRecID())
        {
            recFound = true;
            break;
        }
    }
    if (! recFound)
    {
        report("Map_Connection::get_feature_coords", "invalid feature requested %d", Error::NOTICE, __LINE__, featId);
        return_error_image();
        return;
    }

    co << "Content-Type: application/octet-stream\n\n";
    std::ostringstream msg;
    msg << "COORDLIST," << _id << std::ends;
    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());
    std::string tmp( msg.str());
    co << header << tmp;

    std::ostringstream coords;
    get_coords_for_feature(*fiter, coords);
    coords << std::ends;
    tmp = coords.str();

    report("Map_Connection::get_feature_coords", "coords: %s", Error::DEBUG, __LINE__, tmp.c_str());
    report("Map_Connection::get_feature_coords", "strlen( coords ): %d", Error::DEBUG, __LINE__, tmp.length());

    co << tmp;
    
    co.flush();
}


void Map_Connection::get_multi_feature_coords(const std::string& feature_indices, const std::string &trans_indices)
{
    std::vector<std::string> stringList;
    TransID tran_id;

    report("Map_Connection::get_multi_feature_coords", "feature indices: %s", Error::DEBUG, __LINE__, feature_indices.c_str());
    report("Map_Connection::get_multi_feature_coords", "trans indices: %s", Error::DEBUG, __LINE__, trans_indices.c_str());
    report("Map_Connection::get_multi_feature_coords", "feature list length: %d", Error::DEBUG, __LINE__, _current_feature_list.size());

    co << "Content-Type: application/octet-stream\n\n";
    std::ostringstream msg;
    msg << "MCOORDLST," << _id << std::ends;
    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());
    std::string tmp( msg.str());
    co << header << tmp;

    sutil::split(feature_indices, stringList, ",");
    for ( std::vector<std::string>::const_iterator i = stringList.begin(); i != stringList.end(); i++ )
    {
        int curID = atoi((*i).c_str());
        bool recFound = false;
        for (FeatureList::iterator fiter = _current_feature_list.begin(); fiter != _current_feature_list.end(); fiter++)
        {
            report("Map_Connection::get_multi_feature_coords", "curID: %d; recId: %d", Error::DEBUG, __LINE__, curID, fiter->getRecID());
            if ( curID == fiter->getRecID())
            {
                std::ostringstream coords;
                if (get_coords_for_feature(*fiter, coords))
                {
                    co << " ~ " << curID << " ";
                    coords << std::ends;
                    tmp = coords.str();
                    report("Map_Connection::get_multi_feature_coords", "coords: %s", Error::DEBUG, __LINE__, tmp.c_str());
                    report("Map_Connection::get_multi_feature_coords", "strlen( coords ): %d", Error::DEBUG, __LINE__, tmp.length());
                    co << tmp;
                }
                recFound = true;
                break;
            }
        }
        if (! recFound)
        {
            report("Map_Connection::get_multi_feature_coords", "invalid feature requested %d", Error::NOTICE, __LINE__, curID);
            return_error_image();
            return;
        }
    }

    stringList.erase(stringList.begin(), stringList.end());
    sutil::split(trans_indices, stringList, ",");
    for ( std::vector<std::string>::const_iterator i = stringList.begin(); i != stringList.end(); i++ )
    {
        tran_id = atoi((*i).c_str());
        report("Map_Connection::get_multi_feature_coords", "tran_id: %d", Error::DEBUG, __LINE__, tran_id);
        std::ostringstream coords;
        if (get_tran_coords(tran_id, coords))
        {
            co << " ~ " << tran_id << " ";
            coords << std::ends;
            tmp = coords.str();
            report("Map_Connection::get_multi_feature_coords", "coords: %s", Error::DEBUG, __LINE__, tmp.c_str());
            report("Map_Connection::get_multi_feature_coords", "strlen( coords ): %d", Error::DEBUG, __LINE__, tmp.length());
            co << tmp;
        }
        else
        {
            report("Map_Connection::get_multi_feature_coords", "invalid trans feature requested", Error::NOTICE, __LINE__);
            return_error_image();
            return;
        }
    }
    co.flush();
}

/// this is a testing function
int Map_Connection::evaluate_hull()
{
    Vertex nw;
    Vertex se;
    double dx, dy;
    
    sscanf( get_value( "NWX" ).c_str(), "%lf", &dx );
    sscanf( get_value( "NWY" ).c_str(), "%lf", &dy );

    nw.set( dx, dy );
    
    sscanf( get_value( "SEX" ).c_str(), "%lf", &dx );
    sscanf( get_value( "SEY" ).c_str(), "%lf", &dy );

    se.set( dx, dy );

    if (!imap::setStateCounty(get_value("STATE"), get_value("COUNTY"), mapBase))
    {
        report("Map_Connection::evaluate_hull", "Failed to set state/county", Error::ERROR, __LINE__);
        co << HTTPHTMLHeader() << html() << body() << "no features found" << body() << html() << std::endl; co.flush();
        return false;
    }

    report("Map_Connection::evaluate_hull", "evaluate hull: nw(%f,%f) - se(%f,%f)", Error::DEBUG, __LINE__, nw.x(), nw.y(), se.x(), se.y());

    RectHull * rh = new RectHull( nw, se );
    FeatureList fl;

    if ( create_image(*rh)){
        std::ofstream ofs( "/tmp/map.jpg" );
        ofs.write( (const char *)_image_ptr, _image_size );
        ofs.close();
    }

    co << HTTPHTMLHeader() <<
        html()<<
        body() << "Test images generated in /tmp" << body() << 
        html() <<
        std::endl;
    
    co.flush();

    return 0;
}


int Map_Connection::get_district_polygons()
{
    std::string buf;

    if (!imap::setStateCounty(get_value( "STATE" ), get_value("COUNTY"), mapBase))
    {
        report("Map_Connection::get_district_polygons", "Failed to set state/county", Error::ERROR, __LINE__);
        co << HTTPHTMLHeader() << html() << body() << "no features found" << body() << html() << std::endl; co.flush();
        return 0;
    }

    OvlQuery query;

    query.requestExtended();
    query.requestCombine();
    query.requestState( mapBase->getStateCode() );
    query.requestCounty( mapBase->getCountyCode() );
    query.clearType();
    if (FeatureDef::getSearch(FeatureDef::DistPoly) == true)
        query.requestType( FeatureDef::DistPoly );

    buf = get_value( "DISTRICT-CODE" );
    sutil::uppercase( buf );
    
    std::string dist_id = buf.c_str();

    report("Map_Connection::get_district_polygons", "dist_id %s", Error::DEBUG, __LINE__, dist_id.c_str());
    
    FName fname;
    fname.name = mapBase->getNameManager().getDistrictID( dist_id );

    report("Map_Connection::get_district_polygons", "snapshot", Error::DEBUG, __LINE__);
    
    if ( fname.name == NameMgr::InvalidID ){
        report("Map_Connection::get_district_polygons", "invalid feature name", Error::WARNING, __LINE__);
        co << HTTPHTMLHeader() << html() << body() << "no features found" << body() << html() << std::endl; co.flush();
        return 0;
    }

    fname.prefix = FeatureDef::AnyDirection;
    fname.type   = FeatureDef::AnyNameType;
    fname.suffix = FeatureDef::AnyDirection;

    query.requestName( fname );

    _current_feature_list.erase(_current_feature_list.begin(), _current_feature_list.end());
    
    if (! OvlMgr::getFeatures( query , _current_feature_list )){
        report("Map_Connection::get_district_polygons", "Could not read features", Error::WARNING, __LINE__);
        return_empty_list();
        return 0;
    }

    return_district_polygons( _current_feature_list );

    return 0;
}


void Map_Connection::get_configuration(std::string& clientMode, const std::string& op_id)
{
    ColorTxt colorTxt;
    std::ostringstream msg;
    char header[ 11 ];
    msl::IMapConfig iMapConfig;

//    config->set_op_id_parms(op_id);

    co << "Content-Type: application/octet-stream\n\n";

    msg << "CLIENTCONFIG" << std::ends;
    sprintf( header, "%-10d", msg.str().size());
    std::string tmp( msg.str());
    co << header << tmp;

    colorTxt = config->district_line_color();
    co << "~" << (long)colorTxt.red() << " " << (long)colorTxt.green() << " " << (long)colorTxt.blue();
    co << "~" << config->district_line_width();

    colorTxt = config->wip_line_color();
    co << "~" << (long)colorTxt.red() << " " << (long)colorTxt.green() << " " << (long)colorTxt.blue();
    co << "~" << config->wip_line_width();

    colorTxt = config->wtv_line_color();
    co << "~" << (long)colorTxt.red() << " " << (long)colorTxt.green() << " " << (long)colorTxt.blue();
    co << "~" << config->wtv_line_width();

    colorTxt = config->cmt_line_color();
    co << "~" << (long)colorTxt.red() << " " << (long)colorTxt.green() << " " << (long)colorTxt.blue();
    co << "~" << config->cmt_line_width();

    sutil::uppercase(clientMode);
    report("Map_Connection::get_configuration", "Client is in mode %s, op_id %s", Error::DEBUG, __LINE__, clientMode.c_str(), op_id.c_str());
    if (clientMode == MAPX_MODE) // check for mapx since there are 3 versions of FEDIT
        iMapConfig = config->getMapxConfig();
    else
        iMapConfig = config->getFeditConfig();
    
    co << "~" << iMapConfig.maxPoly;
    co << "~" << iMapConfig.minPoly;
    if (iMapConfig.checkGrid)
        co << "~1";
    else
        co << "~0";

    co.flush();
}

void Map_Connection::get_privs(const std::string& op_id)
{
    std::ostringstream msg;
    char header[ 11 ];

//    config->set_op_id_parms(op_id);
    msl::IMapPrivs iMapPrivs = config->getPrivs();

    co << "Content-Type: application/octet-stream\n\n";

    msg << "CLIENTPRIVS" << std::ends;
    sprintf( header, "%-10d", msg.str().size());
    std::string tmp( msg.str());
    co << header << tmp;

    if (iMapPrivs.modifyDistrict)
        co << "~1";
    else
        co << "~0";
    if (iMapPrivs.deleteDistrict)
        co << "~1";
    else
        co << "~0";

    if (iMapPrivs.addNewWIP)
        co << "~1";
    else
        co << "~0";
    if (iMapPrivs.modifyWIP)
        co << "~1";
    else
        co << "~0";
    if (iMapPrivs.deleteWIP)
        co << "~1";
    else
        co << "~0";

    if (iMapPrivs.submitWIPToWTV)
        co << "~1";
    else
        co << "~0";
    if (iMapPrivs.processWTV)
        co << "~1";
    else
        co << "~0";

    if ( iMapPrivs.default_locate_poly)
        co << "~1";
    else
        co << "~0";

    co.flush();
}

void Map_Connection::get_scale()
{
    std::ostringstream msg;
    char header[ 11 ];
    std::string scaleStr = SCALE_PREFIX;

    co << "Content-Type: application/octet-stream\n\n";

    msg << "SCALE" << std::ends;
    sprintf( header, "%-10d", msg.str().size());
    string tmp( msg.str());
    co << header << tmp;

    double dist = calculate_scale();
    if (dist < 1.0)
    {
        long distForRound = (long) (dist * Vertex::FEET_PER_MILE) ;
        scaleStr += ntos(distForRound) ;
        scaleStr += SCALE_UNIT_FEET ;
    }
    else
    {
        long distForRound = (long) (dist * 10.0) ;
        dist = (double) distForRound / 10.0 ;
        scaleStr += ntos(dist) ;
        scaleStr += SCALE_UNIT_MILES ;
    }
    co << "~" << scaleStr;

    co.flush();
}

double Map_Connection::calculate_scale()
{
    RectHull rh = _hull_manager.get_current_hull();
    return rh[NE].distanceGEO(rh[NW]) / _image_width;
}

void Map_Connection::return_district_polygons( const FeatureList & features )
{
    co << "Content-Type: application/octet-stream\n\n";

    std::ostringstream msg;
    
    msg << "FEATURLIST," << _id << std::ends;

    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());

    std::string tmp( msg.str());
    co << header << tmp;
    
    std::string fname;
    Feature::RecID recID;
    std::ostringstream buf;
    
    for (FeatureList::const_iterator iter = features.begin(); iter != features.end(); iter++)
    {
        mapBase->getNameManager().getFullFeatureNameString((*iter).getMainName(), fname );

            
        if ( fname.length() ){
            buf << "~" << (*iter).getRecID() << "~" << fname << "~";
            report("Map_Connection::return_district_polygons", "fname: %s", Error::DEBUG, __LINE__, fname.c_str());

            if (RingUtil::outerRingFeature(*iter, mapBase->getNameManager()))
                buf << "0";
            else if (RingUtil::innerRingFeature(*iter, mapBase->getNameManager(), recID))
                buf << recID;
/* feature alias' aren't very interesting at first glance...
   
            if ( (*iter).hasNames()){
                FNameList fnl = (*iter).getNames();
                FNameListIndex fnli;
                FNameString fns;
                
                ListLoop( fnl, fnli ){
                    mapBase->getNameManager().getFNameString( *(fnl[ fnli ]), fns );
                    co << " " << fns.name;
                    report("Map_Connection::return_district_polygons", "fname: %s", Error::DEBUG, __LINE__, static_cast<const char *>(*(fns.name));
                }
            }
*/
        }
    }

    buf << std::ends;
    tmp = buf.str();

    co << tmp;
    co.flush();
}

    
int Map_Connection::process_transaction()
{
    std::string transaction_code = get_value( "TRANS-TYPE" );
    std::string transaction_status = get_value( "TRANS-STATUS" );
    std::string district_code = get_value( "DISTRICT" );
    std::string comments = get_value( "COMMENTS" );
    std::string activation_date = get_value( "ACTIVATION-DATE" );
    std::string user_name = get_value( "USER_NAME" );
    int featId = atol( get_value( "FEATURE-INDEX" ).c_str());

    report("Map_Connection::process_transaction", "TRANS-TYPE:    %s", Error::DEBUG, __LINE__, transaction_code.c_str());
    report("Map_Connection::process_transaction", "TRANS-STATUS:  %s", Error::DEBUG, __LINE__, transaction_status.c_str());
    report("Map_Connection::process_transaction", "DISTRICT:      %s", Error::DEBUG, __LINE__, district_code.c_str());
    report("Map_Connection::process_transaction", "COMMENTS:      %s", Error::DEBUG, __LINE__, comments.c_str());
    report("Map_Connection::process_transaction", "FEATURE-INDEX: %d", Error::DEBUG, __LINE__, featId);
    report("Map_Connection::process_transaction", "DATE:          %s", Error::DEBUG, __LINE__, activation_date.c_str());
    report("Map_Connection::process_transaction", "USER:          %s", Error::DEBUG, __LINE__, user_name.c_str());

    TransID tran_id = 0;

    if ( transaction_code == "ADD" && transaction_status == "WORK" ){
        report("Map_Connection::process_transaction", "ADD-WORK", Error::DEBUG, __LINE__);

        TransRecord rec( TransRecord::SOURCE_IMAP, user_name.c_str(), mapBase->getStateCode(), mapBase->getCountyCode() );
        rec.setKind( TransRecord::TRANS_ADDNEW );
        rec.setComments( comments.c_str());
        rec.setStatus( TransRecord::STATUS_WORK );
        rec.setActDate( activation_date.c_str());
        rec.setFName( district_code.c_str());
        
        int num_points = atoi( get_value( "POINTS" ).c_str());

        Polygon polygon;
        char xtag[ 16 ], ytag[ 16 ];
    
        for ( int i = 0; i < num_points; i++ ){
            sprintf( xtag, "X%d", i ); sprintf( ytag, "Y%d", i );
            Vertex v( atof( get_value( xtag ).c_str()), atof( get_value( ytag ).c_str()));
            polygon.insertBack(v);
        }

        TransFeatureData trans_feat;
        trans_feat.clear();
        trans_feat.setType( 8 );
        trans_feat.setSubType( 0 );
        trans_feat.setMainName( district_code.c_str());
        trans_feat.setCoords( polygon );

        msl::TransactionDB db(config->getStateCode());
        
        if (( tran_id = db.createAddTrans( rec, trans_feat )) == 0 ){
            report("Map_Connection::process_transaction", "error creating transaction", Error::ERROR, __LINE__);
        }            
    }
    else if ( transaction_code == "MOD" && transaction_status == "WORK" ){
        report("Map_Connection::process_transaction", "MODIFY-WORK", Error::DEBUG, __LINE__);

        TransRecord rec( TransRecord::SOURCE_IMAP, user_name.c_str(), mapBase->getStateCode(), mapBase->getCountyCode() );
        rec.setKind( TransRecord::TRANS_MODIFY );
        rec.setComments( comments.c_str());
        rec.setStatus( TransRecord::STATUS_WORK );
        rec.setActDate( activation_date.c_str());
        rec.setFName( district_code.c_str());

        int num_points = atoi( get_value( "POINTS" ).c_str());

        Polygon polygon;
        char xtag[ 16 ], ytag[ 16 ];
    
        for ( int i = 0; i < num_points; i++ ){
            sprintf( xtag, "X%d", i ); sprintf( ytag, "Y%d", i );
            Vertex v( atof( get_value( xtag ).c_str()), atof( get_value( ytag ).c_str()));
            polygon.insertBack(v);
        }

        TransFeatureData new_feat;
        new_feat.clear();
        new_feat.setType( 8 );
        new_feat.setSubType( 0 );
        new_feat.setMainName( district_code.c_str());
        new_feat.setCoords( polygon );

        FeatureList::iterator iter = _current_feature_list.begin();
        for (; iter != _current_feature_list.end(); iter++)
        {
            if ( featId == iter->getRecID())
                break;
        }            
        TransFeatureData old_feat(*iter);

        rec.setOrigFeatId(featId);

        msl::TransactionDB db(config->getStateCode());

        if (( tran_id = db.createModTrans( rec, old_feat, new_feat )) == 0 ){
            report("Map_Connection::process_transaction", "error creating modify transaction", Error::ERROR, __LINE__);
        }            
    }
    else if ( transaction_code == "DEL" && transaction_status == "WORK" ){
        report("Map_Connection::process_transaction", "DEL-WORK", Error::DEBUG, __LINE__);
        
        TransRecord rec;
        rec.clear();
        rec.setSource( TransRecord::SOURCE_IMAP );
        rec.setUid( user_name.c_str() );
        rec.setState( mapBase->getStateCode() );
        rec.setCounty( mapBase->getCountyCode() );
        rec.setComments( comments.c_str());
        rec.setActDate( activation_date.c_str());
        rec.setFName( district_code.c_str());

        FeatureList::iterator iter = _current_feature_list.begin();
        for (; iter != _current_feature_list.end(); iter++)
        {
            if ( featId == iter->getRecID())
                break;
        }            
        TransFeatureData trans_feat(*iter);

        msl::TransactionDB db(config->getStateCode());

        if (( tran_id = db.createDelTrans( rec, trans_feat )) == 0 ){
            report("Map_Connection::process_transaction", "error creating delete transaction", Error::ERROR, __LINE__);
        }
    }
    else if ( transaction_status == "SAVE" ){
        report("Map_Connection::process_transaction", "SAVE WORK", Error::DEBUG, __LINE__);

        int num_points = atoi( get_value( "POINTS" ).c_str());

        Polygon polygon;
        char xtag[ 16 ], ytag[ 16 ];
    
        for ( int i = 0; i < num_points; i++ ){
            sprintf( xtag, "X%d", i ); sprintf( ytag, "Y%d", i );
            Vertex v( atof( get_value( xtag ).c_str()), atof( get_value( ytag ).c_str()));
            polygon.insertBack(v);
        }

        tran_id = featId;
        msl::TransactionDB db(config->getStateCode());
        TransRecord rec;
        TransFeatureData origFeat;
        TransFeatureData newFeat;
        newFeat.clear();
        newFeat.setType( 8 );
        newFeat.setSubType( 0 );
        newFeat.setMainName( district_code.c_str());
        newFeat.setCoords( polygon );
        if (db.getTrans(tran_id, rec))
        {
            ulong feat_id = rec.getNewFeatId();
            if (feat_id <= 0)
            {
                feat_id = rec.getOrigFeatId();
            }
            if (db.getTransData(feat_id, origFeat))
            {
                newFeat.setRecID(origFeat.getRecID());
            }
        }

        if ( ! db.saveTrans( tran_id, TransRecord::SOURCE_IMAP, user_name.c_str(), comments.c_str(), newFeat )){
            report("Map_Connection::process_transaction", "error saving work updates", Error::ERROR, __LINE__);
            tran_id = 0;
        }            
        if (rec.getActDate() != activation_date.c_str())
            if (! db.changeActDate(tran_id, activation_date.c_str(), TransRecord::SOURCE_IMAP, user_name.c_str(), comments.c_str()))
                report("Map_Connection::process_transaction", "error changing activation date", Error::ERROR, __LINE__);
    }
    else if ( transaction_status == "SUBMIT" ){
        report("Map_Connection::process_transaction", "SUBMIT", Error::DEBUG, __LINE__);

        tran_id = featId;
        msl::TransactionDB db(config->getStateCode());

        if ( ! db.submitTrans( tran_id, TransRecord::SOURCE_IMAP, user_name.c_str(), comments.c_str())){
            report("Map_Connection::process_transaction", "error submitting tranaction tid(%d)", Error::ERROR, __LINE__, tran_id);
            tran_id = 0;
        }
    }
    else if ( transaction_status == "VERIFY" ){
        report("Map_Connection::process_transaction", "VERIFY", Error::DEBUG, __LINE__);

        tran_id = featId;
        msl::TransactionDB db(config->getStateCode());

        if ( ! db.verifyTrans( tran_id, TransRecord::SOURCE_IMAP, user_name.c_str(), comments.c_str())){
            report("Map_Connection::process_transaction", "error verifying tranaction tid(%d)", Error::ERROR, __LINE__, tran_id);
            tran_id = 0;
        }
    }
    else if ( transaction_status == "REJECT" ){
        report("Map_Connection::process_transaction", "REJECT", Error::DEBUG, __LINE__);

        tran_id = featId;
        msl::TransactionDB db(config->getStateCode());

        if ( ! db.unSubmitTrans( tran_id, TransRecord::SOURCE_IMAP, user_name.c_str(), comments.c_str())){
            report("Map_Connection::process_transaction", "error rejecting transaction tid(%d)", Error::DEBUG, __LINE__, tran_id);
            tran_id = 0;
        }
    }
    else if ( transaction_status == "DELETE" ){
        report("Map_Connection::process_transaction", "DELETE", Error::DEBUG, __LINE__);

        tran_id = featId;
        msl::TransactionDB db(config->getStateCode());

        if ( ! db.deleteTrans( tran_id, TransRecord::SOURCE_IMAP, user_name.c_str(), comments.c_str())){
            report("Map_Connection::process_transaction", "error deleting transaction tid(%d)", Error::DEBUG, __LINE__, tran_id);
            tran_id = 0;
        }
    }
        
    // return transaction id...

    std::ostringstream msg;
    msg << "TRANID," << tran_id << std::ends;
    std::string tmp( msg.str());

    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());

    co << "Content-Type: application/octet-stream\n\n";
    co << header << tmp;
    co.flush();

    return 0;
}




void Map_Connection::draw_error_page( const std::string & msg )
{
    std::string post_action = "http://" + _base->_cgi->getEnvironment().getServerName() + _map_server;
    report("Map_Connection::draw_error_page", "post_action: %s", Error::DEBUG, __LINE__, post_action.c_str());

    co << HTTPHTMLHeader() << HTMLDoctype( HTMLDoctype::eTransitional ) << std::endl;
    co << html() << head() << std::endl;
    co << meta().set( "http-equiv", "Content-Type" ).set( "content", "text/html; charset=utf-8" ) << std::endl;
    co << title( "ERROR" ) << std::endl;
    co << script().set( "type", "text/javascript" ) << std::endl;
    co << "" << std::endl;
    co << "function submit_form( op ){" << std::endl;
    co << "  document.error_form.op_code.value = op;" << std::endl;
    co << "  document.error_form.submit();" << std::endl;
    co << "}" << std::endl;
    co << "" << std::endl;
    co << script() << std::endl;
    co << head();
    co << body();
    co << msg << std::endl;
    co << cgicc::form().set( "name", "error_form" ).set( "method", "post" ).set( "action", post_action ) << std::endl;
    co << input().set( "type", "hidden" ).set( "name", "op_code" ).set( "value", "" ) << std::endl;
    co << input().set( "type", "hidden" ).set( "name", "id" ).set( "value", sutil::ntos( _id )) << std::endl;
    co << input().set( "type", "button" ).set( "value", "Continue" ).set( "onclick", "submit_form('EXIT');" ) << std::endl;
    co << cgicc::form() << std::endl;
    co << body();
    co << html();
    co.flush();
}


int Map_Connection::draw_fedit_screen( bool edit_privs, bool verify_privs )
{
    AccessRecordList access_list;
    msl::IMapDB db(config->getStateCode());

    report("Map_Connection::draw_fedit_screen", "_caller_id: %s", Error::DEBUG, __LINE__, get_value("caller_id").c_str());

    if ( !  db.getAccessRecordList( atol( get_value( "caller_id" ).c_str()), access_list )){
        report("Map_Connection::draw_fedit_screen", "error retrieving access list", Error::WARNING, __LINE__);
        draw_error_page( "No districts authorized for this caller id" );
        return Connection::CONTINUE;
    }

    SysParm aParm;
    std::string applet, javascript;

    report("Map_Connection::draw_fedit_screen", "access_list.size(%d)", Error::DEBUG, __LINE__, access_list.size());
    report("Map_Connection::draw_fedit_screen", "access record: %d, %s, %s, %s", Error::DEBUG, __LINE__,
           access_list[0].getCallerID(), access_list[0].getStateName().c_str(), 
           access_list[0].getCountyName().c_str(), access_list[0].getDistrict().c_str());

    co << HTTPHTMLHeader() << HTMLDoctype( HTMLDoctype::eTransitional ) << std::endl;
    co << html() << head() << std::endl;
    co << meta().set( "http-equiv", "Content-Type" ).set( "content", "text/html; charset=utf-8" ) << std::endl;

    std::string versionText;
    mapcgiVersion::getVersion(versionText);
    adtVersion::getVersion(versionText);
    appVersion::getVersion(versionText);
    comVersion::getVersion(versionText);
    drawVersion::getVersion(versionText);
    geoVersion::getVersion(versionText);
    guiVersion::getVersion(versionText);
    mapVersion::getVersion(versionText);
    modeVersion::getVersion(versionText); // not used by mapcgi, but used by other libraries
    mslVersion::getVersion(versionText);
    msldbVersion::getVersion(versionText);
    nadVersion::getVersion(versionText);
    ovlVersion::getVersion(versionText);
    pssVersion::getVersion(versionText);
    co << "<!-- mapcgi version: " << std::endl << versionText.c_str() << "-->" << std::endl;

    co << title( "Powered by iMap, One Call Concepts, Inc." ) << std::endl;

    string stylesheet( config->web_dir()); stylesheet += "occ.css";
    
    co << cgicc::link().set( "REL", "stylesheet" ).set( "HREF", stylesheet ) << std::endl;

    aParm.getPair(SysParmStr::IMAP_JAVASCRIPT_NAME_PARM);
    javascript = aParm.getValue(DEFAULT_IMAP_JAVASCRIPT_NAME);
    co << script().set( "type", "text/javascript" ).set( "src", config->web_dir() + "help.js" ) << script() << std::endl;
    co << script().set( "type", "text/javascript" ).set( "src", config->web_dir() + "help_data.js" ) << script() << std::endl;
    co << script().set( "type", "text/javascript" ).set( "src", config->web_dir() + javascript.c_str() ) << script() << std::endl;

    co << script().set( "type", "text/javascript" ) << std::endl;
    co << "" << std::endl;
    co << "function logout( op ){" << std::endl;
    co << "  document.imap_form.op_code.value = op;" << std::endl;
    co << "  document.imap_form.submit();" << std::endl;
    co << "}" << std::endl;

    // create a stub function for imenu shutdown
    co << "function check_imenu_shutdown(){ return; }" << std::endl;

    co << script() << std::endl;

    co << head() << std::endl;
    co << body() << std::endl;

    co << table().set( "align", "center" ).set( "valign", "top" ).set( "bgcolor", "silver" ).set( "cellpadding", "2" ).set( "cellspacing", "0" ).set( "border", "1" ) << std::endl;
    co << tbody() << std::endl;
    co << "<tr align=left valign=middle width=100% height=100%>" << std::endl;
    co << cgicc::form().set( "name", "imap_form" ) << std::endl;

    co << input().set( "type", "hidden" ).set( "name", "op_code" ).set( "value", "" ) << std::endl;
    co << input().set( "type", "hidden" ).set( "name", "id" ).set( "value", sutil::ntos( _id )) << std::endl;

    co << "<td align=left valign=middle width=100% height=100%>" << std::endl;

    co << "<table width=100% height=100% cellpadding=0 cellspacing=0>" << std::endl;

    // create counyt_state_dist lookup list in javascript to enable
    // dynamically created state and district select lists.
    
    co << script().set( "type", "text/javascript" ) << std::endl;
    co << "var county_state_dist_list = new Array(";

    std::string tmpStr;
    AccessRecordList::const_iterator last_county = access_list.end(); last_county--;

    for ( AccessRecordList::const_iterator i = access_list.begin(); i != last_county; i++ )
    {
        tmpStr = i->getStateName();
        lrtrim(tmpStr);
        report("Map_Connection::draw_fedit_screen", "county: %s state: %s district: %s", Error::DEBUG, __LINE__,
               i->getCountyName().c_str(), tmpStr.c_str(), i->getDistrict().c_str());

        co << "new Array(\"" << i->getCountyName() << "\",\"" << tmpStr.c_str() << "\",\"" << i->getDistrict() << "\")," << std::endl;
    }

    tmpStr = last_county->getStateName();
    lrtrim(tmpStr);
    co << "new Array(\"" << last_county->getCountyName() << "\",\"" << tmpStr << "\",\"" << last_county->getDistrict() << "\"));" << std::endl;
    co << script() << std::endl;

    co << "<tr align=left >" << std::endl;    
    co << "<td valign=bottom>" << std::endl;
    string occ_logo( config->web_dir()); occ_logo += "images/ocb.png";
    co << img().set( "title", "One Call Concepts, Inc." ).set( "width", "129" ).set("height", "30" ).set( "src", occ_logo ) << std::endl;
    co << "<br></td></tr>" << std::endl;
    
    co << "<tr align=center >" << std::endl;    
    // create the county select list
    co << "<td>&nbsp;&nbsp;County:" << std::endl;
    co << cgicc::select().set( "name", "county" ).set( "onchange", "build_dist_state_list(this.form.dist_code,this.form.state,this.form.county,county_state_dist_list);" ) << std::endl;
    County_Access_List cal;
    get_county_list( access_list, cal );

    co << option( "" ).set( "value", "" );
    
    for ( County_Access_List::const_iterator i = cal.begin(); i != cal.end(); i++ ){
        co << option( *i ).set( "value", *i );
    }
    co << std::endl << cgicc::select() << std::endl;
    
    // create the state select list
    co << "&nbsp;State:" << std::endl;
    co << cgicc::select().set( "name", "state" ).set( "onchange", "build_dist_list(this.form.dist_code,this.form.state,this.form.county,county_state_dist_list);" ) << std::endl;
    co << std::endl << cgicc::select() << std::endl;

    // create the district code select list
    co << "&nbsp;District:" << std::endl;
    co << cgicc::select().set( "name", "dist_code" );
    co << cgicc::select() << std::endl;
    co << "&nbsp;" << input().set( "type", "button" ).set( "name", "dist_search" ).set( "value", "Go!" ).set( "onclick", "do_dist_search();" ) << std::endl;
    co << "&nbsp;" << input().set( "type", "button" ).set( "name", "exit" ).set( "value", "Exit" ).set( "onclick", "logout('EXIT');" ) << std::endl;
    
    co << "</td></tr>" << std::endl;
    co << "</table>" << std::endl;
    co << "</td>" << std::endl;

    co << cgicc::form() << std::endl;

    co << "</tr>" << std::endl;

    co << "<tr align=center valign=middle width=100% height=100%>" << std::endl;
    co << "<td align=center valign=middle width=100% height=100%>" << std::endl;

    string codebase( config->web_dir()); codebase += "applets";

    aParm.getPair(SysParmStr::IMAP_APPLET_NAME_PARM);
    applet = aParm.getValue(DEFAULT_IMAP_APPLET_NAME);
    co << "<applet code=\"OCCPad.class\" codebase=\"" << codebase << "\" name=imap height=585 width=700 archive=\"" << applet.c_str() << "\">" << std::endl;
    co << param().set( "name", "url" ).set( "value", _map_server ) << std::endl;
    co << param().set( "name", "state" ) << std::endl;
    co << param().set( "name", "county" ) << std::endl;
    co << param().set( "name", "place" ) << std::endl;
    co << param().set( "name", "address" ) << std::endl;
    co << param().set( "name", "street" ) << std::endl;
    co << param().set( "name", "intersection" ) << std::endl;
    co << param().set( "name", "coord_type" ) << std::endl;
    co << param().set( "name", "nad83" ).set( "value", "true" ) << std::endl;
    co << param().set( "name", "lat" ) << std::endl;
    co << param().set( "name", "lon" ) << std::endl;
    co << param().set( "name", "lat_deg" ) << std::endl;
    co << param().set( "name", "lat_min" ) << std::endl;
    co << param().set( "name", "lat_sec" ) << std::endl;
    co << param().set( "name", "lon_deg" ) << std::endl;
    co << param().set( "name", "lon_min" ) << std::endl;
    co << param().set( "name", "lon_sec" ) << std::endl;
    co << param().set( "name", "northing" ) << std::endl;
    co << param().set( "name", "easting" ) << std::endl;
    co << param().set( "name", "zone" ) << std::endl;
    co << param().set( "name", "user_name" ).set( "value", get_value( "email" )) << std::endl;

    report("Map_Connection::draw_fedit_screen", "email %s", Error::DEBUG, __LINE__, get_value("email").c_str());
    
    if ( edit_privs ){
        if ( verify_privs ){
            co << param().set( "name", "mode" ).set( "value", FEDIT_ALL_MODE ) << std::endl;
        }
        else {
            co << param().set( "name", "mode" ).set( "value", FEDIT_MODE ) << std::endl;
        }
    }
    else {
        if ( verify_privs ){
            co << param().set( "name", "mode" ).set( "value", FEDIT_VERIFY_MODE ) << std::endl;
        }
        else {
            co << param().set( "name", "mode" ).set( "value", FEDIT_VIEW_MODE ) << std::endl;
        }
    }
    
    co << param().set( "name", "district_code" ).set( "value", "" ) << std::endl;
    co << param().set( "name", "map_id" ).set( "value", sutil::ntos( _id )) << std::endl;
    co << param().set( "name", "op_id" ).set( "value", get_value("op_id")) << std::endl;
    co << param().set( "name", "hotspot" ).set( "value", "8" ) << std::endl;
    co << "</applet>" << std::endl;
    co << "</td></tr>" << std::endl;
    co << tbody() << table() << std::endl;
    co << body() << std::endl;
    co << html() << std::endl;
    co.flush();

    return CONTINUE;
}


int Map_Connection::return_trans_records( const TransRecordList & trl )
{
    co << "Content-Type: application/octet-stream\n\n";

    std::ostringstream msg;
    
    msg << "TRANSLIST," << _id << std::ends;

    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());

    std::string tmp( msg.str());
    co << header << tmp;

    std::ostringstream buf;
    TransFeatureData tfd;
    msl::TransactionDB db(config->getStateCode());
    
    for ( TransRecordList::const_iterator i = trl.begin(); i != trl.end(); i++ ){
        ulong feat_id = i->getOrigFeatId() > 0 ? i->getOrigFeatId() : i->getNewFeatId();
        
        if ( db.getTransData( feat_id, tfd )){
            buf << "~" << i->getTransID() << "~" << i->getKindStr() << "~" << i->getFName() << "~" << tfd.getRecID() << "~" << i->getUid() << "~" << i->getActDate() << "~" << i->getComments();
        }
        else {
            buf << "~" << i->getTransID() << "~" << i->getKindStr() << "~" << i->getFName() << "~" << "0000000000" << "~" << i->getUid() << "~" << i->getActDate() << "~" << i->getComments();
        }
    }
        
    buf << std::ends;
    tmp = buf.str();

    co << tmp;
    co.flush();
    return 0;
}


int Map_Connection::return_empty_list()
{
    co << "Content-Type: application/octet-stream\n\n";

    std::ostringstream msg;
    
    msg << "ERROR," << _id << std::ends;

    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());

    std::string tmp( msg.str());
    co << header << tmp;
    co.flush();
    return 0;
}


int Map_Connection::get_tran_records( TransRecord::TransStatus status )
{

    if (!imap::setStateCounty(get_value("STATE"), get_value("COUNTY"), mapBase))
    {
        report("Map_Connection::get_tran_records", "Failed to set state/county", Error::ERROR, __LINE__);
        co << HTTPHTMLHeader() << html() << body() << "no features found" << body() << html() << std::endl; co.flush();
        return 0;
    }

    std::string user_name = get_value( "USER_NAME" );
    std::string districtCode = get_value("DISTRICT-CODE");

    report("Map_Connection::get_tran_records", "District code %s user: %s", Error::DEBUG, __LINE__, districtCode.c_str(), user_name.c_str());

    TransQuery trans_query;
    trans_query.clear();
    trans_query.setState( mapBase->getStateCode() );
    trans_query.setCounty( mapBase->getCountyCode() );
    trans_query.setUser( user_name.c_str());
    trans_query.setMainName(districtCode);
    trans_query.setStatus( status );

    _current_wip_list.clear();

    msl::TransactionDB db(config->getStateCode());
    
    if ( db.queryTrans( trans_query, _current_wip_list ) > 0 ){
        report("Map_Connection::get_tran_records", "db.queryTrans found %d records", Error::DEBUG, __LINE__, _current_wip_list.size());
        return_trans_records( _current_wip_list );
    }
    else {
        report("Map_Connection::get_tran_records", "no records found", Error::DEBUG, __LINE__);        
        return_empty_list();
    }

    return 0;
}


int Map_Connection::get_wip_records()
{
    return get_tran_records( TransRecord::STATUS_WORK );
}


int Map_Connection::get_wtv_records()
{
    return get_tran_records( TransRecord::STATUS_SUBMITTED );
}


int Map_Connection::get_committed_records()
{
    return get_tran_records( TransRecord::STATUS_VERIFIED );
}

int Map_Connection::get_tran_image()
{
    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end()); // erase any highlights
    TransID tran_id = atol( get_value( "FEATURE_INDEX" ).c_str());
    
    report("Map_Connection::get_tran_image", "trans Id: %d", Error::DEBUG, __LINE__, tran_id);

    TransRecord rec;
    TransFeatureData tfd;
    msl::TransactionDB db(config->getStateCode());

    if ( db.getTrans( tran_id, rec )){
        ulong feat_id = rec.getNewFeatId();

        if ( feat_id <= 0 ){
            feat_id = rec.getOrigFeatId();
        }
    
        if ( db.getTransData( feat_id, tfd )){
            RectHull tfd_hull;
            tfd.getHull( tfd_hull );
        
            const RectHull & rh = _hull_manager.reset( const_cast< RectHull & >( tfd_hull ));
            create_image(rh);
            transmit_image();
            return 0;
        }
    }

    return_error_image();
    return 0;
}


bool Map_Connection::get_tran_coords(const TransID& tran_id, std::ostringstream& coords)
{
    bool found = false;
    TransRecord rec;
    msl::TransactionDB db(config->getStateCode());
    TransFeatureData tfd;

    if ( db.getTrans( tran_id, rec ))
    {
        ulong feat_id = rec.getNewFeatId();

        if ( feat_id <= 0 )
        {
            feat_id = rec.getOrigFeatId();
        }

        if ( db.getTransData( feat_id, tfd ))
        {
            Geometric g;
            tfd.getGeometric( g );

            coords.precision( 8 );

            if ( g.type() == MSLenum::POLYGON )
            {
                found = true;
                int points = g.size();
                report("Map_Connection::get_tran_coords", "points = %d", Error::DEBUG, __LINE__, points);

                Polygon * p = static_cast< Polygon * >( g.getShape() );

                coords << p->size();

                for (int i = 0; i < p->size(); i++)
                    coords << " " << (*p)[i].x() << " " << (*p)[i].y();
            }

            coords << std::ends;

        }
    }
    return found;
}

int Map_Connection::get_tran_polygon()
{
    TransID tran_id = atol( get_value( "FEATURE_INDEX" ).c_str());
    report("Map_Connection::get_tran_polygon", "tran_id: %d", Error::DEBUG, __LINE__, tran_id);
    std::ostringstream coords;

    if (get_tran_coords(tran_id, coords))
    {
        std::ostringstream msg;

        co << "Content-Type: application/octet-stream\n\n";
        msg << "COORDLIST," << _id << std::ends;
        char header[ 11 ];
        sprintf( header, "%-10d", msg.str().size());
        std::string tmp( msg.str());
        co << header << tmp;
        co << coords.str();
        co.flush();
        return 0;
    }
     
    return_empty_list();
    return 0;
}

void Map_Connection::handle_ping_request()
{
    co << HTTPPlainHeader() << "PING" << std::endl;
}

void Map_Connection::get_msg_coord_image(int msgNumber, int revNumber, double scale_pct, const std::string& recIDs)
{
    _highlight_features.erase(_highlight_features.begin(), _highlight_features.end());

    PrismDBI prismDBI(imap::DB_USER_NAME); // Note the non-namespace, need the environment varables

    MessageCoordList coords;
    FeatureList* distPolyList = NULL;

    if (prismDBI.getMessageCoordHull(msgNumber, revNumber, coords))
    {
        message_t msg;
        state_t state;
        county_t county;

        // Get the state and county
        if (prismDBI.getMessage(msgNumber, revNumber, msg) < DBI_SUCCESS)
        {
            report("Map_Connection::get_msg_coord_image", "Failed to retrieve message %d rev %d from the DB", Error::WARNING, __LINE__, msgNumber, revNumber);
            return_error_image(true);
            return;
        }
        if (prismDBI.getState(msg.state, state) < DBI_SUCCESS)
        {
            report("Map_Connection::get_msg_coord_image", "Failed to lookup state number for state \"%s\" from the DB", Error::WARNING, __LINE__, msg.state);
            return_error_image(true);
            return;
        }
        if (prismDBI.getCounty(msg.county, state.state_no, county) < DBI_SUCCESS)
        {
            report("Map_Connection::get_msg_coord_image", "Failed to lookup county number for county \"%s\" in state \"%s\" (%d) from the DB", Error::WARNING, __LINE__, msg.county, msg.state, state.state_no);
            return_error_image(true);
            return;
        }
        mapBase->setStateCode(state.state_no);
        mapBase->setCountyCode(county.county_no);
        report("Map_Connection::get_msg_coord_image", "Creating view image for message %d rev %d state %d county %d recIDs: \"%s\"", Error::DEBUG, __LINE__, msgNumber, revNumber, state.state_no, county.county_no, recIDs.c_str());

        MessageCoordList::const_iterator iter = coords.begin();
        RectHull rh = Vertex((*iter).lon, (*iter).lat);
        for (iter++; iter != coords.end(); iter++)
            rh += Vertex((*iter).lon, (*iter).lat);
        coords.erase(coords.begin(), coords.end());
        report("Map_Connection::get_msg_coord_image", "RectHull NW(%f, %f) SE(%f %f)", Error::DEBUG, __LINE__, rh[NW].x(), rh[NW].y(), rh[SE].x(), rh[SE].y());

        // Enlarge the hull so the locate poly isnt' on the edges
        if (prismDBI.getMessageCoordNoHull(msgNumber, revNumber, coords))
        {
            bool polyIsCircle = false;
            // We need to determine if the poly is a circle
            if (coords.size() == 2)
            {
                for (iter = coords.begin(); iter != coords.end(); iter++)
                {
                    if ((*iter).lat == 99)
                    {
                        polyIsCircle = true;
                        break;
                    }
                }
            }

            Feature locateFeat;
            if (!polyIsCircle)
            {
                Polygon locatePoly;
                for (iter = coords.begin(); iter != coords.end(); iter++)
                {
                    report("Map_Connection::get_msg_coord_image", "Adding locate point (%f, %f)", Error::DEBUG, __LINE__, (*iter).lon, (*iter).lat);
                    locatePoly.insertBack(Vertex((*iter).lon, (*iter).lat));
                }
                locateFeat.setGeometric(locatePoly);
                locateFeat.setHull(locatePoly.getRectHull());
            }
            else
            {
                Circle locateCir;
                for (iter = coords.begin(); iter != coords.end(); iter++)
                {
                    // Is this the center or the radius
                    if ((*iter).lat != 99)
                        locateCir.setCenter(Vertex((*iter).lon, (*iter).lat));
                    else
                        locateCir.setRadius((*iter).lon);
                }
                locateFeat.setGeometric(locateCir);
                locateFeat.setHull(locateCir.getRectHull());
            }
            _highlight_features.push_back(locateFeat);
        }

        Hull_Manager::frame_nicely(rh);
        if (scale_pct != 0)
            rh *= ( 100.0 + scale_pct ) / 100.0;

        if (recIDs.length() > 0)
        {
            report("Map_Connection::get_msg_coord_image", "Adding district poly records", Error::DEBUG, __LINE__);
            distPolyList = new FeatureList();
            getDistrictPolygonList(mapBase->getStateCode(), mapBase->getCountyCode(), recIDs, *distPolyList);
            report("Map_Connection::get_msg_coord_image", "%d district polygons found", Error::DEBUG, __LINE__, distPolyList->size());
        }

        create_image(rh, distPolyList);
        if (distPolyList)
        {
            distPolyList->clear();
            delete distPolyList;
        }
        transmit_image(false);
    }
    else
    {
        report("Map_Connection::get_msg_coord_image", "Failed to lookup message coording for msg_no %d rev_no %d from the DB", Error::WARNING, __LINE__, msgNumber, revNumber);
        return_error_image(true);
    }
    return;
}

void Map_Connection::export_map_data(const std::string& districtCode, const std::string& state, const std::string& emailAddress, const std::string& counties, const std::string& outputFormat, const std::string& compressFormat, const std::string& statePath, const std::string& includeBaseMap)
{
    std::string versionText;
    std::string resultString;

    report("Map_Connection::export_map_data", "Starting for district %s state %s counties %s", Error::DEBUG, __LINE__, districtCode.c_str(), state.c_str(), counties.c_str());

    mapcgiVersion::getVersion(versionText);
    adtVersion::getVersion(versionText);
    appVersion::getVersion(versionText);
    comVersion::getVersion(versionText);
    drawVersion::getVersion(versionText);
    geoVersion::getVersion(versionText);
    guiVersion::getVersion(versionText);
    mapVersion::getVersion(versionText);
    modeVersion::getVersion(versionText); // not used by mapcgi, but used by other libraries
    mslVersion::getVersion(versionText);
    msldbVersion::getVersion(versionText);
    nadVersion::getVersion(versionText);
    ovlVersion::getVersion(versionText);
    pssVersion::getVersion(versionText);

    co << HTTPHTMLHeader() << std::endl;
    co << "<!-- mapcgi version: " << versionText << " -->" << std::endl;

    report("Map_Connection::export_map_data", "Starting for district %s state %s counties %s", Error::DEBUG, __LINE__, districtCode.c_str(), state.c_str(), counties.c_str());

    ExportDistrictData::doExportMapData(mapBase->getOverlayPath(), districtCode, state, emailAddress, counties, outputFormat, compressFormat, statePath, includeBaseMap, resultString);

    co << resultString << std::endl;
    co.flush();

    return;
}

void Map_Connection::get_visible_grids()
{
    if (_hull_manager.get_initialized())
    {
        int x, y, z;
        double dx, dy;
        RectHull rh;
        OvlQuery query;
        FeatureList gList, tList;
        std::ostringstream msg;
        std::string gridName;
        const Shape* shape;
        DrawSetupInfo info;

        // Get the parameters
        sscanf(get_value("NWX").c_str(), "%lf", &dx);
        sscanf(get_value("NWY").c_str(), "%lf", &dy);
        rh.setCorner(NW, dx, dy);

        sscanf(get_value("SEX").c_str(), "%lf", &dx);
        sscanf(get_value("SEY").c_str(), "%lf", &dy);
        rh.setCorner(SE, dx, dy);

        report("Map_Connection::get_visible_grids", "(%i, %i, %i) NW(%f, %f) SE(%f, %f)", Error::DEBUG, __LINE__, x, y, z, rh[NW].x(), rh[NW].y(), rh[SE].x(), rh[SE].y());

        co << "Content-Type: application/octet-stream\n\n";
        msg << "GRIDLIST," << _id << std::ends;
        char header[ 11 ];
        sprintf( header, "%-10d", msg.str().size());
        string tmp( msg.str());
        co << header << tmp;

        // Need to calculate zoom level
        setupImageInfo(info, rh);

        // If not plotting grids, don't return data to the client
        if (info.ftypeSelected[FeatureDef::Grid] == false)
        {
            co << "~0~";
            co.flush();
            return;
        }

        // Get the grids
        query.requestState(mapBase->getStateCode());
        query.requestCounty(mapBase->getCountyCode());
        query.requestHull(rh);
        query.requestType(FeatureDef::Grid);
        // enforce zooming
        query.requestDisplayLevel(PageDraw::calculateDisplayLevel(info.scaleMagnification)) ;
        query.requestGeneralize();
        query.requestCombine();

        if (! OvlMgr::getFeatures(query, gList))
        {
            co << "~0~";
            co.flush();
            return;
        }

        // Get rid of copied features
        if (OvlUtil::removeCopies(gList, tList))
            tList.erase(tList.begin(), tList.end());
        report("Map_Connection::get_visible_grids", "Returning %d grids", Error::DEBUG, __LINE__, gList.size());

        co << "~" << gList.size() << "~";
        for (FeatureList::const_iterator fiter = gList.begin(); fiter != gList.end(); fiter++)
        {
            mapBase->getNameManager().getFullFeatureNameString((*fiter).getMainName(), gridName);
            if (gridName.length())
            {
                if ((*fiter).getGeometric().size() > 0)
                {
                    shape = (*fiter).getGeometric().getShape();
                    co << gridName << "~" << (*fiter).getGeometric().size() << "~";
                    for (int i = 0; i < shape->size(); i++)
                        co << (*shape)[i].x() << "~" << (*shape)[i].y() << "~";
                }
            }
        }
        co.flush();

    }
    return;
}

void Map_Connection::get_zone()
{
    bool validRequest = true;
    std::ostringstream msg;

    if (!imap::setStateCounty(get_value("STATE"), get_value( "COUNTY" ), mapBase))
        validRequest = false;

    co << "Content-Type: application/octet-stream\n\n";
    msg << "ZONEINFO," << _id << std::ends;
    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());
    string tmp(msg.str());
    co << header << tmp;
    if (!validRequest)
    {
        co << "~0~0~";
    }
    else
    {
        co << "~" << ZoneData::getSpcsZone83(mapBase->getStateCode(), mapBase->getCountyCode());
        co << "~" << ZoneData::getUtmZone(mapBase->getStateCode(), mapBase->getCountyCode());
        co << "~";
    }
    co.flush();

}

void Map_Connection::inflate_locate()
{
    Vertex v;
    double lat, lon;
    Polygon reqPoly;
    std::string parmName;
    std::ostringstream msg;
    std::ostringstream poly_buf;

    int pointCount = atoi(get_value("POINT_COUNT").c_str());
    for (int i = 0; i < pointCount; i++)
    {
        parmName = "LAT";
        parmName += ntos(static_cast<long>(i));
        sscanf(get_value(parmName.c_str()).c_str(), "%lf", &lat);
        parmName = "LON";
        parmName += ntos(static_cast<long>(i));
        sscanf(get_value(parmName.c_str()).c_str(), "%lf", &lon);
        reqPoly.insertBack(Vertex(lat, lon));
    }

    Locate locate(mapBase->getStateCode(), mapBase->getCountyCode());
    locate.initLocate(reqPoly, true);

    co << "Content-Type: application/octet-stream\n\n";
    msg << "L_INFLATED," << _id << std::ends;
    char header[ 11 ];
    sprintf( header, "%-10d", msg.str().size());
    string tmp(msg.str());
    co << header << tmp;

    if (failedPrimarySearch)
        co << "Y~";
    else
        co << "N~";
    const Polygon* poly = static_cast<const Polygon*>(locate.getShape());
    create_locate_poly_string(*poly, poly_buf);
    co << poly_buf.str();

    co.flush();
}


void Map_Connection::get_recID_coords(const std::string& state, const std::string& county, const std::string& recIDs)
{
    FeatureList distPolyList;
    std::string tmp;

    report("MapConnection::get_recID_coords", "Collecting coordinates for state \"%s\" county \"%s\" RecID list \"%s\"", Error::DEBUG, __LINE__, state.c_str(), county.c_str(), recIDs.c_str());

    co << "Content-Type: application/octet-stream\n\n";

    if (recIDs.length() < 1)
    {
        report("Map_Connection::get_recID_coords", "No recIDs requested", Error::ERROR, __LINE__);
        co << "~0~";
        co.flush();
        return;

    }

    if (!imap::setStateCounty(state, county, mapBase))
    {
        report("Map_Connection::get_recID_coords", "Failed to set state/county", Error::ERROR, __LINE__);
        co << "~0~";
        co.flush();
        return;
    }

    getDistrictPolygonList(mapBase->getStateCode(), mapBase->getCountyCode(), recIDs, distPolyList);

    report("Map_Connection::get_recID_coords", "%d district polygons found", Error::DEBUG, __LINE__, distPolyList.size());
    if (distPolyList.size() < 1)
    {
        report("Map_Connection::get_recID_coords", "No district polygons found", Error::NOTICE, __LINE__);
        co << "~0~";
        co.flush();
        return;
    }

    co << "~" << distPolyList.size() << "~";
    for (FeatureList::const_iterator iter = distPolyList.begin(); iter != distPolyList.end(); iter++)
    {
        std::ostringstream coords;

        get_coords_for_feature(*iter, coords);
        coords << std::ends;
        tmp = coords.str();
        report("Map_Connection::get_recID_coords", "coords (length %d): %s", Error::DEBUG, __LINE__, tmp.length(), tmp.c_str());
        co << tmp << "~";
    }
    co.flush();

    distPolyList.clear();
    return;

}

/**
 * This method creates a shutdown dialog and updates the Page_Displayer so it can be shown to the user.
 *
 * @param   seconds_left is the number of seconds remaining before the server is terminated.
 */
void Map_Connection::start_shutdown( time_t seconds_left )
{
    errRptD(( "start_shutdown, id: %ld time: %ld", _id, seconds_left ));

    assert( ! shutdown_dialog_ ); assert( getenv( "CONFIGPATH" ));

    shutdown_dialog_ = new Shutdown_Dialog( config->imap_server(), getenv( "CONFIGPATH" ), seconds_left, _id );
    shutdown_message_ = "SHUTDOWN," + sutil::ntos( shutdown_dialog_->width()) + "," + sutil::ntos( shutdown_dialog_->height()) + ",";
}


/**
 * This method updates the time remaining until the server is shutdown in the Shutdown Dialog.
 *
 * @param   seconds_left is the number of seconds remaining before the server is terminated.
 */
void Map_Connection::update_countdown_timer( time_t seconds_left )
{
    assert( shutdown_dialog_ );

    shutdown_dialog_->update_counter( seconds_left );
}

// known op codes
// 
// XPRISM-SEARCH
//     conduct a feature search in the same order as Xprism (address/intersection/trsq/grid)
// STREET-SEARCH
//     conduct an address/street/intersection search
// LOCATE-POLYGON
//
// GET-FEATURE-COORDS
//     
// GET-MULTI-FEATURE-COORDS
// GET-DIST-FEATURE
//    return a district feature given a trans id
// GET-MULTI-DIST-FEATURES
// GET-PAGE
//     return a map image for the given page coordinagtes
// GET-FEATURE
//     return a map image centered around the feature give the index from a previous feature search.
// GET-DIST-POLY-CONF
//     return the fedit application parameters
// EVALUATE-HULL
//     return a map image for the given NW/SE decimal degree lat/lon coordinates
// TRSQ-SEARCH
//     conduct a trsq search
// LL-SEARCH
//     conduct a lat/lon search
// PROCESS-TRANS
//     apply the given transaction to the transaction database
// FEDIT
//     set the connection in mapx mode and return the fedit user screen
// FEDIT-VERIFY
// FEDIT-VIEW
// GET-WIP
//     return work in process
// GET-WORK-TO-VERIFY
//     return work to verify
// GET-TRAN-IMAGE
//     return a map image centered around a feature given a transaction id
// GET-TRAN-COORDS
//     return the feature coordinates for a given transaction id
// GET-SCALE
//     return the scale value of the current hull
// SCALE-HULL
//     return a map image from the given NW/SE decimal degree lat/lon coordinates that has
//     been scaled by the given percentage.
// GET-MSG-COORD-IMAGE
//     returns a map image for a polygon in the Prism message_coord table
// EXPORT-MAP-DATA
//     exports MSL District Polygon data and makes available for download
// GET-GRIDS
//     Gets the visible grids for the given page coordinates
// GET-ZONE
//     Gets the SPCS and UTM zone for the given state & county
// P_INFLATE
//     Creates a locate from the provided polygon (which may or may not inflate)
// GET-RECID-COORDS
//     Gets coords for a list of record IDs
// EXIT
//     redirect to the login URL.
int Map_Connection::process_request()
{
    std::string op_code = get_value( "op_code" );

    errRptI(( "============> Beginging op_code/id = %s/%ld", op_code.c_str(), _id));

    try {
        if ( shutdown_dialog_ ){
            Request_Handler_Status rhs = shutdown_dialog_->handle_request( op_code, *this );

            if ( rhs.first ){
                return rhs.second;
            }
        }

        if ( op_code == "XPRISM-SEARCH" ){
            int street_matches = 0, grid_matches = 0;
            
            if ( shutdown_dialog_ ) send_shutdown_ = true;

            if ( ll_search(true)){
                transmit_image();
            }
            else if (( street_matches = street_search(true)) == 1 ){
                transmit_image();
            }
            else if ( street_matches == 0 && ( grid_matches = trsq_search(true)) > 0 ){
                transmit_image();
            }
            else if ( street_matches == 0 ){
                return_error_image();
            }
        }
        else if ( op_code == "STREET-SEARCH" ){
            int street_matches = 0;
            
            if ( shutdown_dialog_ ) send_shutdown_ = true;

            if (( street_matches = street_search(false)) == 1 ){
                transmit_image();
            }
            else if ( street_matches == 0 ){
                return_error_image();
            }
        }
        else if ( op_code == "TRSQ-SEARCH" ){
            if ( shutdown_dialog_ ) send_shutdown_ = true;

            if ( trsq_search(false) > 0 ){
                transmit_image();
            }
            else {
                return_error_image();
            }
        }
        else if ( op_code == "LL-SEARCH" ){
            if ( shutdown_dialog_ ) send_shutdown_ = true;

            if ( ll_search(false)){
                transmit_image();
            }
            else {
                return_error_image();
            }
        }
        else if ( op_code == "LOCATE-POLYGON" ){
            transmit_locate_polygon();
        }
        else if ( op_code == "GET-FEATURE" ){
            get_feature( atol( get_value( "feature_index" ).c_str()));
        }
        else if ( op_code == "GET-FEATURE-COORDS" ){
            get_feature_coords( atol( get_value( "feature_index" ).c_str()));
        }
        else if ( op_code == "GET-MULTI-FEATURE-COORDS" ){
            std::string feature_indices = get_value("feature_indices");
            std::string trans_indices = get_value("trans_indices");
            get_multi_feature_coords(feature_indices, trans_indices);
        }
        else if ( op_code == "GET-DIST-FEATURE" ){
            get_district_feature( atol( get_value( "feature_index" ).c_str()));
        }
        else if ( op_code == "GET-MULTI-DIST-FEATURES" )
        {
            std::string feature_indices = get_value("feature_indices");
            std::string trans_indices = get_value("trans_indices");
            get_multi_district_features(feature_indices, trans_indices);
        }
        else if ( op_code == "GET-PAGE" ){
            get_image_from_page();
        }
        else if ( op_code == "GET-FEATURES" ){
            get_district_polygons();
        }
        else if ( op_code == "GET-CONFIGURATION") {
            std::string clientMode = get_value("mode");
            std::string op_id = get_value("op_id");
            get_configuration(clientMode, op_id);
        }
        else if ( op_code == "GET-PRIVS") {
            std::string op_id = get_value("op_id");
            get_privs(op_id);
        }
        else if ( op_code == "GET-SCALE") {
            get_scale();
        }
        else if ( op_code == "EVALUATE-HULL" ){
            evaluate_hull();
        }
        else if ( op_code == "SCALE-HULL" ){
            if ( shutdown_dialog_ ) send_shutdown_ = true;

            if ( ! get_value( "PREFS" ).empty()){
                user_prefs_->write_user_prefs( _email, get_value( "PREFS" ));
                user_prefs_->apply_user_prefs( *mapBase );
            }
            if ( get_scaled_image()){
                transmit_image();
            }
            else {
                return_error_image();
            }
        }
        else if ( op_code == "IDENTIFY-FEATURES" ) { // Identify features around a given point
             get_identify_features();
        }
        else if ( op_code == "IDENTIFY-IMAGE" ) { // Identify features around a given point
            if ( shutdown_dialog_ ) send_shutdown_ = true;

            if (get_identify_image())
                transmit_image();
            else
                return_error_image();
        }
        else if ( op_code == "PROCESS-TRANS" ){
            process_transaction();
        }
        else if ( op_code == "FEDIT" ){
            return draw_fedit_screen( true, false );
        }
        else if ( op_code == "FEDIT-VERIFY" ){
            return draw_fedit_screen( false, true );
        }
        else if ( op_code == "FEDIT-VIEW" ){
            return draw_fedit_screen( false, false );
        }
        else if ( op_code == "GET-WIP" ){ // list of all work in progress (labels)
            return get_wip_records();
        }
        else if ( op_code == "GET-WORK-TO-VERIFY" ){ // list of all submitted work (labels)
            return get_wtv_records();
        }
        else if ( op_code == "GET-COMMITTED" ){ // list of all committed transactions (labels)
            return get_committed_records();
        }
        else if ( op_code == "GET-TRAN-IMAGE" ){ // image for a specific piece of submitted work
            return get_tran_image();
        }
        else if ( op_code == "GET-TRAN-COORDS" ){ // polygon for a specific piece of work
            return get_tran_polygon();
        }
        else if (op_code == "GET-MSG-COORD-IMAGE") // image for existing message
        {
            int msgNumber = atoi(get_value("msg_no").c_str());
            int revNumber = atoi(get_value("rev_no").c_str());
            double scale_pct = atof(get_value("zoom_pct").c_str());
            std::string recIDs = get_value("recids"); // optional
            get_msg_coord_image(msgNumber, revNumber, scale_pct, recIDs);
            errRptI(( "============> Completed op_code = %s", op_code.c_str()));
            return TERMINATE;
        }
        else if (op_code == "EXPORT-MAP-DATA") // export District Polygons
        {
            std::string districtCode = get_value("district_code");
            std::string state = get_value("state");
            std::string emailAddress = get_value("email");
            std::string counties = get_value("counties");
            std::string outputFormat = get_value("outputFormat");
            std::string compressFormat = get_value("compressFormat");
            std::string statePath = get_value("statePath");
            std::string includeBaseMap = get_value("includeBaseMap");
            export_map_data(districtCode, state, emailAddress, counties, outputFormat, compressFormat, statePath, includeBaseMap);
            errRptI(( "============> Completed op_code = %s", op_code.c_str()));
            return TERMINATE;
        }
        else if (op_code == "GET-GRIDS")
        {
            get_visible_grids();
        }
        else if (op_code == "GET-ZONE")
        {
            get_zone();
        }
        else if (op_code == "L_INFLATE")
        {
            inflate_locate();
        }
        else if ( op_code == "EXIT" ){
            draw_exit_page();
            errRptI(( "============> Completed op_code = %s", op_code.c_str()));
            return TERMINATE;
        }
        else if ( op_code == "PING" ){
            handle_ping_request();
            errRptI(( "============> Completed op_code = %s", op_code.c_str()));
            return TERMINATE;
        }
        else if ( op_code == "TERMINATE" ){
            return_error_image();
            errRptI(( "============> Completed op_code = %s", op_code.c_str()));
            return TERMINATE;
        }
        else if ( user_prefs_->handle_request( op_code, _id, _base->_cout )){
            errRptI(( "============> Completed op_code = %s", op_code.c_str()));
        }
        else if ( op_code == "APPLY_PREFS" ){
            user_prefs_->write_user_prefs( _email, get_value( "PREFS" ));
            user_prefs_->apply_user_prefs( *mapBase );
            
            errRptI(( "prefs= %s", get_value( "PREFS" ).c_str()));
            return_error_image();
            errRptI(( "============> Completed op_code = %s", op_code.c_str()));
        }
        else if (op_code == "GET-RECID-COORDS")
        {
            std::string state = get_value("state");
            std::string county = get_value("county");
            std::string recIDs = get_value("recids");
            get_recID_coords(state, county, recIDs);
            errRptI(( "============> Completed op_code = %s", op_code.c_str()));
            return TERMINATE;
        }
        else {
            unknown_request( op_code );
        }
    }
    catch ( const char * msg ){
        report("Map_Connection::process_request", msg, Error::NOTICE, __LINE__);
//        return_server_error();
        return_error_image();
    }

    errRptI(( "============> Completed op_code = %s", op_code.c_str()));
    return Connection::CONTINUE;
}
